# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    A test class to check that the translation process yields a circuit identical to the native one for the
    different backends supported
"""

import unittest
import os
import numpy as np

from tangelo.linq import Gate, Circuit
from tangelo.linq.gate import PARAMETERIZED_GATES
import tangelo.linq.translator as translator
from tangelo.helpers.utils import installed_backends

path_data = os.path.dirname(os.path.realpath(__file__)) + '/data'

gates = [Gate("H", 2), Gate("CNOT", 1, control=0), Gate("CNOT", 2, control=1), Gate("Y", 0), Gate("S", 0)]
abs_circ = Circuit(gates) + Circuit([Gate("RX", 1, parameter=2.)])
multi_controlled_gates = [Gate("X", 0), Gate("X", 1), Gate("CX", target=2, control=[0, 1])]
abs_multi_circ = Circuit(multi_controlled_gates)
init_gates = [Gate('H', 0), Gate('X', 1), Gate('H', 2)]
one_qubit_gate_names = ["H", "X", "Y", "Z", "S", "T", "RX", "RY", "RZ", "PHASE"]
one_qubit_gates = [Gate(name, target=0) if name not in PARAMETERIZED_GATES else Gate(name, target=0, parameter=0.5)
                   for name in one_qubit_gate_names]
one_qubit_gates += [Gate(name, target=1) if name not in PARAMETERIZED_GATES else Gate(name, target=1, parameter=0.2)
                    for name in one_qubit_gate_names]
two_qubit_gate_names = ["CNOT", "CX", "CY", "CZ", "CPHASE", "CRZ"]
two_qubit_gates = [Gate(name, target=1, control=0) if name not in PARAMETERIZED_GATES
                   else Gate(name, target=1, control=0, parameter=0.5) for name in two_qubit_gate_names]
swap_gates = [Gate('SWAP', target=[1, 0]), Gate('CSWAP', target=[1, 2], control=0)]
big_circuit = Circuit(init_gates + one_qubit_gates + two_qubit_gates + swap_gates + [Gate('XX', [0, 1], parameter=0.5)])

references = [0., 0.38205142 ** 2, 0., 0.59500984 ** 2, 0., 0.38205142 ** 2, 0., 0.59500984 ** 2]
references_multi = [0., 0., 0., 0., 0., 0., 0., 1.]
reference_big_lsq = [-0.29022980 + 0.20684454j, -0.34400320 + 0.12534970j,  0.21316957 + 0.23442923j,
                      0.15939614 + 0.15293439j, -0.36723378 + 0.29031223j, -0.04807413 + 0.0797184j,
                     -0.37427732 + 0.41885117j, -0.05511766 + 0.20825736j]
reference_big_msq = [-0.29022979 + 0.20684454j, -0.36723376 + 0.29031221j,  0.21316958 + 0.23442923j,
                     -0.37427729 + 0.41885117j, -0.34400321 + 0.12534970j, -0.04807414 + 0.07971841j,
                      0.15939615 + 0.15293440j, -0.05511766 + 0.20825737j]

abs_circ_mixed = Circuit(gates) + Circuit([Gate("RX", 1, parameter=1.5), Gate("MEASURE", 0)])


class TestTranslation(unittest.TestCase):

    @unittest.skipIf("qulacs" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_qulacs(self):
        """
            Compares the results of a simulation with Qulacs using a qulacs circuit directly
            VS one obtained through translation from an abstract format
        """
        import qulacs

        # Generates the qulacs circuit by translating from the abstract one
        translated_circuit = translator.translate_qulacs(abs_circ)

        # Run the simulation
        state1 = qulacs.QuantumState(abs_circ.width)
        translated_circuit.update_quantum_state(state1)

        # Directly define the same circuit through qulacs
        # NB: this includes convention fixes for some parametrized rotation gates (-theta instead of theta)
        qulacs_circuit = qulacs.QuantumCircuit(3)
        qulacs_circuit.add_H_gate(2)
        qulacs_circuit.add_CNOT_gate(0, 1)
        qulacs_circuit.add_CNOT_gate(1, 2)
        qulacs_circuit.add_Y_gate(0)
        qulacs_circuit.add_S_gate(0)
        qulacs_circuit.add_RX_gate(1, -2.)  # Convention: sign of theta

        # Run the simulation
        state2 = qulacs.QuantumState(abs_circ.width)
        qulacs_circuit.update_quantum_state(state2)

        # Assert that both simulations returned the same state vector
        np.testing.assert_array_equal(state1.get_vector(), state2.get_vector())

        # Generates the qulacs circuit by translating from the abstract one
        translated_circuit = translator.translate_qulacs(abs_multi_circ)

        # Run the simulation
        state1 = qulacs.QuantumState(abs_multi_circ.width)
        translated_circuit.update_quantum_state(state1)

        # Directly define the same circuit through qulacs
        # NB: this includes convention fixes for some parametrized rotation gates (-theta instead of theta)
        qulacs_circuit = qulacs.QuantumCircuit(3)
        qulacs_circuit.add_X_gate(0)
        qulacs_circuit.add_X_gate(1)
        mat_gate = qulacs.gate.to_matrix_gate(qulacs.gate.X(2))
        mat_gate.add_control_qubit(0, 1)
        mat_gate.add_control_qubit(1, 1)
        qulacs_circuit.add_gate(mat_gate)

        # Run the simulation
        state2 = qulacs.QuantumState(abs_multi_circ.width)
        qulacs_circuit.update_quantum_state(state2)

        # Assert that both simulations returned the same state vector
        np.testing.assert_array_equal(state1.get_vector(), state2.get_vector())

        # Test that the translated circuit reports the same result for all cross-supported gates
        translated_circuit = translator.translate_qulacs(big_circuit)

        # Run the simulation
        state1 = qulacs.QuantumState(big_circuit.width)
        translated_circuit.update_quantum_state(state1)
        np.testing.assert_array_almost_equal(state1.get_vector(), reference_big_msq, decimal=6)

    @unittest.skipIf("qiskit" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_qiskit(self):
        """
            Compares the results of a simulation with Qulacs using a qulacs circuit directly
            VS one obtained through translation from an abstract format
        """

        import qiskit

        # Generate the qiskit circuit by translating from the abstract one and print it
        translated_circuit = translator.translate_qiskit(abs_circ)

        # Generate the qiskit circuit directly and print it
        circ = qiskit.QuantumCircuit(3, 3)
        circ.h(2)
        circ.cx(0, 1)
        circ.cx(1, 2)
        circ.y(0)
        circ.s(0)
        circ.rx(2., 1)

        # Simulate both circuits, assert state vectors are equal
        qiskit_simulator = qiskit.Aer.get_backend("aer_simulator", method='statevector')
        translated_circuit = qiskit.transpile(translated_circuit, qiskit_simulator)
        circ = qiskit.transpile(circ, qiskit_simulator)
        translated_circuit.save_statevector()
        circ.save_statevector()

        sim_results = qiskit_simulator.run(translated_circuit).result()
        v1 = sim_results.get_statevector(translated_circuit)

        sim_results = qiskit_simulator.run(circ).result()
        v2 = sim_results.get_statevector(circ)

        np.testing.assert_array_equal(v1, v2)

        # Return error when attempting to use qiskit with multiple controls
        self.assertRaises(ValueError, translator.translate_qiskit, abs_multi_circ)

        # Generate the qiskit circuit by translating from the abstract one and print it
        translated_circuit = translator.translate_qiskit(big_circuit)
        # Simulate both circuits, assert state vectors are equal
        qiskit_simulator = qiskit.Aer.get_backend("aer_simulator", method='statevector')
        translated_circuit = qiskit.transpile(translated_circuit, qiskit_simulator)
        translated_circuit.save_statevector()
        sim_results = qiskit_simulator.run(translated_circuit).result()
        np.testing.assert_array_almost_equal(sim_results.get_statevector(translated_circuit), reference_big_msq, decimal=6)

    @unittest.skipIf("cirq" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_cirq(self):
        """
            Compares the results of a simulation with cirq using a cirq circuit directly
            VS one obtained through translation from an abstract format
        """

        import cirq

        # Generate the qiskit circuit by translating from the abstract one and print it
        translated_circuit = translator.translate_cirq(abs_circ)

        # Generate the cirq circuit directly and print it
        qubit_labels = cirq.LineQubit.range(3)
        circ = cirq.Circuit()
        circ.append(cirq.H(qubit_labels[2]))
        circ.append(cirq.CNOT(qubit_labels[0], qubit_labels[1]))
        circ.append(cirq.CNOT(qubit_labels[1], qubit_labels[2]))
        circ.append(cirq.Y(qubit_labels[0]))
        circ.append(cirq.S(qubit_labels[0]))
        gate_rx = cirq.rx(2.)
        circ.append(gate_rx(qubit_labels[1]))

        # Simulate both circuits, assert state vectors are equal
        cirq_simulator = cirq.Simulator()

        job_sim = cirq_simulator.simulate(circ)
        v1 = job_sim.final_state_vector

        job_sim = cirq_simulator.simulate(translated_circuit)
        v2 = job_sim.final_state_vector

        np.testing.assert_array_equal(v1, v2)

        translated_circuit = translator.translate_cirq(abs_multi_circ)
        circ = cirq.Circuit()
        circ.append(cirq.X(qubit_labels[0]))
        circ.append(cirq.X(qubit_labels[1]))
        next_gate = cirq.X.controlled(num_controls=2)
        circ.append(next_gate(qubit_labels[0], qubit_labels[1], qubit_labels[2]))

        job_sim = cirq_simulator.simulate(circ)
        v1 = job_sim.final_state_vector

        job_sim = cirq_simulator.simulate(translated_circuit)
        v2 = job_sim.final_state_vector

        np.testing.assert_array_equal(v1, v2)

        # Test that translated circuit is correct for all cross-supported gates
        translated_circuit = translator.translate_cirq(big_circuit)
        job_sim = cirq_simulator.simulate(translated_circuit)
        np.testing.assert_array_almost_equal(job_sim.final_state_vector, reference_big_lsq, decimal=6)

    @unittest.skipIf("qdk" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_qdk(self):
        """ Compares the frequencies computed by the QDK/Q# shot-based simulator to the theoretical ones """

        # Generate the qdk circuit by translating from the abstract one and print it
        translated_circuit = translator.translate_qsharp(abs_circ)
        print(translated_circuit)

        # Write to file
        with open('tmp_circuit.qs', 'w+') as f_out:
            f_out.write(translated_circuit)

        # Compile all qsharp files found in directory and import the qsharp operation
        import qsharp
        qsharp.reload()
        from MyNamespace import EstimateFrequencies

        # Simulate, return frequencies
        n_shots = 10**5
        probabilities = EstimateFrequencies.simulate(nQubits=abs_circ.width, nShots=n_shots)
        print("Q# frequency estimation with {0} samples: \n {1}".format(n_shots, probabilities))

        # Compares with theoretical probabilities obtained through a statevector simulator
        np.testing.assert_almost_equal(np.array(probabilities), np.array(references), 2)

        # Generate the qdk circuit by translating from the abstract one and print it
        translated_circuit = translator.translate_qsharp(abs_multi_circ)
        print(translated_circuit)

        # Write to file
        with open('tmp_circuit.qs', 'w+') as f_out:
            f_out.write(translated_circuit)

        # Compile all qsharp files found in directory and import the qsharp operation
        import qsharp
        qsharp.reload()
        from MyNamespace import EstimateFrequencies

        # Simulate, return frequencies
        n_shots = 10**4
        probabilities = EstimateFrequencies.simulate(nQubits=abs_multi_circ.width, nShots=n_shots)
        print("Q# frequency estimation with {0} samples: \n {1}".format(n_shots, probabilities))

        # Compares with theoretical probabilities obtained through a statevector simulator
        np.testing.assert_almost_equal(np.array(probabilities), np.array(references_multi), 2)

    @unittest.skipIf("projectq" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_projectq(self):
        """ Compares state vector of native ProjectQ circuit against translated one """
        from projectq.ops import All, Measure, H, CX, Y, S, Rx
        from projectq import MainEngine

        translated_circuit = translator.translate_projectq(abs_circ)
        instructions = translated_circuit.split("\n")

        eng = MainEngine()
        Qureg = eng.allocate_qureg(abs_circ.width)
        for instruction in instructions[abs_circ.width:]:
            exec(instruction)
        eng.flush()
        _, v1 = eng.backend.cheat()
        All(Measure) | Qureg
        eng.flush()

        # # Native ProjectQ circuit
        eng2 = MainEngine()
        Qureg2 = eng2.allocate_qureg(3)
        H | Qureg2[2]
        CX | (Qureg2[0], Qureg2[1])
        CX | (Qureg2[1], Qureg2[2])
        Y | Qureg2[0]
        S | Qureg2[0]
        Rx(2.0) | Qureg2[1]
        eng2.flush()
        _, v2 = eng2.backend.cheat()
        All(Measure) | Qureg2
        eng2.flush()

        # Compare statevectors
        np.testing.assert_array_equal(v1, v2)

    @unittest.skipIf("projectq" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_projectq2abs(self):
        """ Compares state vector of native ProjectQ circuit against one translated to abstract and then
        ran with projectQ """

        # Compares original abstract circuit to the one obtained by translating to projectQ and back to abstract
        projectq_circ = translator.translate_projectq(abs_circ)
        abs_circ2 = translator._translate_projectq2abs(projectq_circ)
        assert(abs_circ.__str__() == abs_circ2.__str__())

        # Inverse test: assume input is a ProjectQ circuit such as the output of the CommandPrinter engine
        with open(f"{path_data}/projectq_circuit.txt", 'r') as pq_circ_file:
            pq_circ1 = pq_circ_file.read()
            abs_circ1 = translator._translate_projectq2abs(pq_circ1)
            pq_circ2 = translator.translate_projectq(abs_circ1)

            # This package does not generate final measurements and deallocations, so that simulation can retrieve
            # the statevector beforehand. We append them manually for the sake of this test.
            for i in range(abs_circ1.width):
                pq_circ2 += f"Deallocate | Qureg[{i}]\n"
            assert(pq_circ1 == pq_circ2)

    @unittest.skipIf("qiskit" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_abs2openqasm(self):
        """
            Compares raw openQASM string as generated by Qiskit at the time this test is written, to the one
            this package generates using translate_qiskit and the Qiskit built-in QASM translation

            This test failing implies that either Qiskit QASM output has changed or that translate_qiskit fails
            (the latter has its own tests)
        """
        openqasm_circuit1 = '''OPENQASM 2.0;\ninclude "qelib1.inc";\nqreg q[3];\ncreg c[3];\nh q[2];\ncx q[0],q[1];\ncx '''\
                            '''q[1],q[2];\ny q[0];\ns q[0];\nrx(1.5) q[1];\nmeasure q[0] -> c[0];\n'''
        openqasm_circuit2 = translator.translate_openqasm(abs_circ_mixed)
        print(openqasm_circuit2)

        # For DEBUG later, if the behavior of Qiskit changes
        # import difflib
        # output_list = [li for li in difflib.ndiff(openqasm_circuit1, openqasm_circuit2) if li[0] != ' ']
        # print(output_list)

        assert(openqasm_circuit1 == openqasm_circuit2)

    @unittest.skipIf("qiskit" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_openqasm2abs(self):
        """ Translate from abstract format to openQASM and back, compare with original. """
        openqasm_str = translator.translate_openqasm(abs_circ_mixed)
        abs_circ_mixed2 = translator._translate_openqasm2abs(openqasm_str)

        # Two abstract circuits are identical if and only if they have identical string representations
        assert(abs_circ_mixed.__str__() == abs_circ_mixed2.__str__())

    def test_json_ionq(self):
        """ Translate abstract format to IonQ JSON format """

        abs_gates = [Gate("X", 0), Gate("X", 1), Gate("RX", 0, parameter=1.5707963267948966),
                     Gate("H", 2), Gate("CNOT", target=1, control=0), Gate("RZ", 2, parameter=12.566170614359173)]
        abs_circ_ionq = Circuit(abs_gates)
        json_ionq_circ = translator.translate_json_ionq(abs_circ_ionq)

        ref_circuit = {'circuit': [{'gate': 'x', 'target': 0},
                                   {'gate': 'x', 'target': 1},
                                   {'gate': 'rx', 'rotation': 1.5707963267948966, 'target': 0},
                                   {'gate': 'h', 'target': 2},
                                   {'control': 0, 'gate': 'cnot', 'target': 1},
                                   {'gate': 'rz', 'rotation': 12.566170614359173, 'target': 2}],
                       'qubits': 3}

        assert(json_ionq_circ == ref_circuit)

    def test_translate_ionq_inverse(self):
        """ Test that inverse of T and S circuits for ionQ return Tdag and Sdag after translation """

        # Generate [Gate("Tdag", 0), Gate("Sdag", 0)] equivalent, and its hardcoded inverse
        circ = Circuit([Gate("PHASE", 0, parameter=-np.pi/4), Gate("PHASE", 0, parameter=-np.pi/2)])
        inverse_circ = Circuit([Gate("S", 0), Gate("T", 0)])

        ionq_circ_inverse = translator.translate_json_ionq(circ.inverse())
        ionq_inverse_circ = translator.translate_json_ionq(inverse_circ)
        ionq_circ = translator.translate_json_ionq(circ)

        ionq_ref = {'qubits': 1, 'circuit': [{'gate': 'ti', 'target': 0}, {'gate': 'si', 'target': 0}]}
        self.assertTrue(ionq_inverse_circ == ionq_circ_inverse)
        self.assertTrue(ionq_circ == ionq_ref)

    @unittest.skipIf("braket" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_braket(self):
        """
            Compares the results of a simulation with Braket local simulator using a Braket circuit directly
            VS one obtained through translation from an abstract format
        """
        from braket.circuits import Circuit as BraketCircuit
        from braket.devices import LocalSimulator as BraketLocalSimulator

        # Generate the braket circuit by translating from the abstract one and print it
        translated_circuit = translator.translate_braket(abs_circ)
        print(translated_circuit)

        # Equivalent native braket circuit
        circ = BraketCircuit()
        circ.h(2)
        circ.cnot(0, 1)
        circ.cnot(1, 2)
        circ.y(0)
        circ.s(0)
        circ.rx(1, 2.)
        print(circ)

        # Simulate both circuits on Braket LocalSimulator, assert state vectors are equal
        circ.state_vector()
        translated_circuit.state_vector()

        device = BraketLocalSimulator()
        circ_result = device.run(circ, shots=0).result()
        translated_result = device.run(translated_circuit, shots=0).result()

        np.testing.assert_array_equal(circ_result.values[0], translated_result.values[0])

        # Return error when attempting to use braket with multiple controls
        self.assertRaises(ValueError, translator.translate_braket, abs_multi_circ)

        # Test that circuit is correct for all cross-supported gates
        translated_circuit = translator.translate_braket(big_circuit)
        translated_circuit.state_vector()
        translated_result = device.run(translated_circuit, shots=0).result()
        np.testing.assert_array_almost_equal(translated_result.values[0], reference_big_lsq, decimal=6)

    @unittest.skipIf("qiskit" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_unsupported_gate(self):
        """ Must return an error if a gate is not supported for the target backend """

        circ = Circuit([Gate("Potato", 0)])
        self.assertRaises(ValueError, translator.translate_qiskit, circ)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import math

from tangelo.toolboxes.ansatz_generator.adapt_ansatz import ADAPTAnsatz
from tangelo.toolboxes.operators import FermionOperator
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping

f_op = FermionOperator("2^ 3^ 0 1") - FermionOperator("0^ 1^ 2 3")
qu_op = fermion_to_qubit_mapping(f_op, "jw")
qu_op.terms = {term: math.copysign(1.0, coeff.imag) for term, coeff in qu_op.terms.items()}


class ADAPTAnsatzTest(unittest.TestCase):

    def test_adaptansatz_init(self):
        """Verify behavior of ADAPTAnsatz class."""

        ansatz = ADAPTAnsatz(n_spinorbitals=4, n_electrons=2)
        ansatz.build_circuit()

    def test_adaptansatz_adding(self):
        """Verify operator addition behavior of ADAPTAnsatz class."""

        ansatz = ADAPTAnsatz(n_spinorbitals=4, n_electrons=2)
        ansatz.build_circuit()

        ansatz.add_operator(qu_op)

        self.assertEqual(ansatz.n_var_params, 1)
        self.assertEqual(ansatz._n_terms_operators, [8])

    def test_adaptansatz_set_var_params(self):
        """Verify variational parameter tuning behavior of ADAPTAnsatz class."""

        ansatz = ADAPTAnsatz(n_spinorbitals=4, n_electrons=2)
        ansatz.build_circuit()

        ansatz.add_operator(qu_op)

        ansatz.set_var_params([1.999])
        self.assertEqual(ansatz.var_params, [1.999])

        with self.assertRaises(ValueError):
            ansatz.set_var_params([1.999, 2.999])


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing datastructures for interfacing with this package
functionalities.
"""

import copy
from dataclasses import dataclass, field

import numpy as np
from pyscf import gto, scf, ao2mo, symm, lib
import openfermion
import openfermion.ops.representations as reps
from openfermion.chem.molecular_data import spinorb_from_spatial
from openfermion.ops.representations.interaction_operator import get_active_space_integrals as of_get_active_space_integrals

from tangelo.toolboxes.molecular_computation.frozen_orbitals import get_frozen_core
from tangelo.toolboxes.qubit_mappings.mapping_transform import get_fermion_operator


def atom_string_to_list(atom_string):
    """Convert atom coordinate string (typically stored in text files) into a
    list/tuple representation suitable for openfermion.MolecularData.
    """

    geometry = []
    for line in atom_string.split("\n"):
        data = line.split()
        if len(data) == 4:
            atom = data[0]
            coordinates = (float(data[1]), float(data[2]), float(data[3]))
            geometry += [(atom, coordinates)]
    return geometry


def molecule_to_secondquantizedmolecule(mol, basis_set="sto-3g", frozen_orbitals=None):
    """Function to convert a Molecule into a SecondQuantizedMolecule.

    Args:
        mol (Molecule): Self-explanatory.
        basis_set (string): String representing the basis set.
        frozen_orbitals (int or list of int): Number of MOs or MOs indexes to
            freeze.

    Returns:
        SecondQuantizedMolecule: Mean-field data structure for a molecule.
    """

    converted_mol = SecondQuantizedMolecule(mol.xyz, mol.q, mol.spin,
                                            basis=basis_set,
                                            frozen_orbitals=frozen_orbitals,
                                            symmetry=mol.symmetry)
    return converted_mol


@dataclass
class Molecule:
    """Custom datastructure to store information about a Molecule. This contains
    only physical information.

    Attributes:
        xyz (array-like or string): Nested array-like structure with elements
            and coordinates (ex:[ ["H", (0., 0., 0.)], ...]). Can also be a
            multi-line string.
        q (float): Total charge.
        spin (int): Absolute difference between alpha and beta electron number.
        n_atom (int): Self-explanatory.
        n_electrons (int): Self-explanatory.
        n_min_orbitals (int): Number of orbitals with a minimal basis set.

    Properties:
        elements (list): List of all elements in the molecule.
        coords (array of float): N x 3 coordinates matrix.
    """
    xyz: list or str
    q: int = 0
    spin: int = 0

    # Defined in __post_init__.
    n_atoms: int = field(init=False)
    n_electrons: int = field(init=False)
    n_min_orbitals: int = field(init=False)

    def __post_init__(self):
        mol = self.to_pyscf()
        self.n_atoms = mol.natm
        self.n_electrons = mol.nelectron
        self.n_min_orbitals = mol.nao_nr()

    @property
    def elements(self):
        return [self.xyz[i][0] for i in range(self.n_atoms)]

    @property
    def coords(self):
        return np.array([self.xyz[i][1] for i in range(self.n_atoms)])

    def to_pyscf(self, basis="CRENBL", symmetry=False, ecp=None):
        """Method to return a pyscf.gto.Mole object.

        Args:
            basis (string): Basis set.
            symmetry (bool): Flag to turn symmetry on
            ecp (dict): Dictionary with ecp definition for each atom e.g. {"Cu": "crenbl"}

        Returns:
            pyscf.gto.Mole: PySCF compatible object.
        """

        mol = gto.Mole(atom=self.xyz)
        mol.basis = basis
        mol.charge = self.q
        mol.spin = self.spin
        mol.symmetry = symmetry
        mol.ecp = ecp if ecp else dict()
        mol.build()

        self.xyz = list()
        for sym, xyz in mol._atom:
            self.xyz += [tuple([sym, tuple([x*lib.parameters.BOHR for x in xyz])])]

        return mol

    def to_file(self, filename, format=None):
        """Write molecule geometry to filename in specified format

        Args:
            filename (str): The name of the file to output the geometry.
            format (str): The output type of "raw", "xyz", or "zmat". If None, will be inferred by the filename
        """
        mol = self.to_pyscf()
        mol.tofile(filename, format)

    def to_openfermion(self, basis="sto-3g"):
        """Method to return a openfermion.MolecularData object.

        Args:
            basis (string): Basis set.

        Returns:
            openfermion.MolecularData: Openfermion compatible object.
        """

        return openfermion.MolecularData(self.xyz, basis, self.spin+1, self.q)


@dataclass
class SecondQuantizedMolecule(Molecule):
    """Custom datastructure to store information about a mean field derived
    from a molecule. This class inherits from Molecule and add a number of
    attributes defined by the second quantization.

    Attributes:
        basis (string): Basis set.
        ecp (dict): The effective core potential (ecp) for any atoms in the molecule.
            e.g. {"C": "crenbl"} use CRENBL ecp for Carbon atoms.
        symmetry (bool or str): Whether to use symmetry in RHF or ROHF calculation.
            Can also specify point group using pyscf allowed string.
            e.g. "Dooh", "D2h", "C2v", ...
        mf_energy (float): Mean-field energy (RHF or ROHF energy depending
            on the spin).
        mo_energies (list of float): Molecular orbital energies.
        mo_occ (list of float): Molecular orbital occupancies (between 0.
            and 2.).
        mean_field (pyscf.scf): Mean-field object (used by CCSD and FCI).
        n_mos (int): Number of molecular orbitals with a given basis set.
        n_sos (int): Number of spin-orbitals with a given basis set.
        active_occupied (list of int): Occupied molecular orbital indexes
            that are considered active.
        frozen_occupied (list of int): Occupied molecular orbital indexes
            that are considered frozen.
        active_virtual (list of int): Virtual molecular orbital indexes
            that are considered active.
        frozen_virtual (list of int): Virtual molecular orbital indexes
            that are considered frozen.
        fermionic_hamiltonian (FermionOperator): Self-explanatory.

    Methods:
        freeze_mos: Change frozen orbitals attributes. It can be done inplace
            or not.

    Properties:
        n_active_electrons (int): Difference between number of total
            electrons and number of electrons in frozen orbitals.
        n_active_sos (int): Number of active spin-orbitals.
        n_active_mos (int): Number of active molecular orbitals.
        frozen_mos (list or None): Frozen MOs indexes.
        actives_mos (list): Active MOs indexes.
    """
    basis: str = "sto-3g"
    ecp: dict = field(default_factory=dict)
    symmetry: bool = False
    frozen_orbitals: list or int = field(default="frozen_core", repr=False)

    # Defined in __post_init__.
    mf_energy: float = field(init=False)
    mo_energies: list = field(init=False)
    mo_occ: list = field(init=False)

    mean_field: scf = field(init=False)

    n_mos: int = field(init=False)
    n_sos: int = field(init=False)

    active_occupied: list = field(init=False)
    frozen_occupied: list = field(init=False)
    active_virtual: list = field(init=False)
    frozen_virtual: list = field(init=False)

    def __post_init__(self):
        super().__post_init__()
        self._compute_mean_field()
        self.freeze_mos(self.frozen_orbitals)

    @property
    def n_active_electrons(self):
        return int(sum([self.mo_occ[i] for i in self.active_occupied]))

    @property
    def n_active_sos(self):
        return 2*len(self.active_mos)

    @property
    def n_active_mos(self):
        return len(self.active_mos)

    @property
    def fermionic_hamiltonian(self):
        return self._get_fermionic_hamiltonian()

    @property
    def frozen_mos(self):
        """This property returns MOs indexes for the frozen orbitals. It was
        written to take into account if one of the two possibilities (occ or
        virt) is None. In fact, list + None, None + list or None + None return
        an error. An empty list cannot be sent because PySCF mp2 returns
        "IndexError: list index out of range".

        Returns:
            list: MOs indexes frozen (occupied + virtual).
        """
        if self.frozen_occupied and self.frozen_virtual:
            return self.frozen_occupied + self.frozen_virtual
        elif self.frozen_occupied:
            return self.frozen_occupied
        elif self.frozen_virtual:
            return self.frozen_virtual
        else:
            return None

    @property
    def active_mos(self):
        """This property returns MOs indexes for the active orbitals.

        Returns:
            list: MOs indexes that are active (occupied + virtual).
        """
        return self.active_occupied + self.active_virtual

    @property
    def active_spin(self):
        """This property returns the spin of the active space.

        Returns:
            int: n_alpha - n_beta electrons of the active occupied orbital space.
        """
        return sum([self.mo_occ[i] == 1 for i in self.active_occupied])

    def _compute_mean_field(self):
        """Computes the mean-field for the molecule. It supports open-shell
        mean-field calculation through openfermionpyscf. Depending on the
        molecule spin, it does a restricted or a restriction open-shell
        Hartree-Fock calculation.

        It is also used for defining attributes related to the mean-field
        (mf_energy, mo_energies, mo_occ, n_mos and n_sos).
        """

        molecule = self.to_pyscf(self.basis, self.symmetry, self.ecp)

        self.mean_field = scf.RHF(molecule)
        self.mean_field.verbose = 0
        self.mean_field.kernel()

        if self.symmetry:
            self.mo_symm_ids = list(symm.label_orb_symm(self.mean_field.mol, self.mean_field.mol.irrep_id,
                                                        self.mean_field.mol.symm_orb, self.mean_field.mo_coeff))
            irrep_map = {i: s for s, i in zip(molecule.irrep_name, molecule.irrep_id)}
            self.mo_symm_labels = [irrep_map[i] for i in self.mo_symm_ids]
        else:
            self.mo_symm_ids = None
            self.mo_symm_labels = None

        self.mf_energy = self.mean_field.e_tot
        self.mo_energies = self.mean_field.mo_energy
        self.mo_occ = self.mean_field.mo_occ

        self.n_mos = molecule.nao_nr()
        self.n_sos = 2*self.n_mos

    def _get_fermionic_hamiltonian(self, mo_coeff=None):
        """This method returns the fermionic hamiltonian. It written to take
        into account calls for this function is without argument, and attributes
        are parsed into it.

        Args:
            mo_coeff (array): The molecular orbital coefficients to use to generate the integrals.

        Returns:
            FermionOperator: Self-explanatory.
        """

        core_constant, one_body_integrals, two_body_integrals = self.get_active_space_integrals(mo_coeff)

        one_body_coefficients, two_body_coefficients = spinorb_from_spatial(one_body_integrals, two_body_integrals)

        molecular_hamiltonian = reps.InteractionOperator(core_constant, one_body_coefficients, 1 / 2 * two_body_coefficients)

        return get_fermion_operator(molecular_hamiltonian)

    def _convert_frozen_orbitals(self, frozen_orbitals):
        """This method converts an int or a list of frozen_orbitals into four
        categories:
        - Active and occupied MOs;
        - Active and virtual MOs;
        - Frozen and occupied MOs;
        - Frozen and virtual MOs.
        Each of them are list with MOs indexes (first one is 0). Note that they
        are MOs labelled, not spin-orbitals (MOs * 2) indexes.

        Args:
            frozen_orbitals (int or list of int): Number of MOs or MOs indexes
                to freeze.

        Returns:
            list: Nested list of active occupied, frozen occupied, active
                virtual and frozen virtual orbital indexes.
        """

        if frozen_orbitals == "frozen_core":
            frozen_orbitals = get_frozen_core(self.to_pyscf(self.basis)) if not self.ecp else 0
        elif frozen_orbitals is None:
            frozen_orbitals = 0

        # First case: frozen_orbitals is an int.
        # The first n MOs are frozen.
        if isinstance(frozen_orbitals, int):
            frozen_orbitals = list(range(frozen_orbitals))
        # Second case: frozen_orbitals is a list of int.
        # All MOs with indexes in this list are frozen (first MO is 0, second is 1, ...).
        # Everything else raise an exception.
        elif not (isinstance(frozen_orbitals, list) and all(isinstance(_, int) for _ in frozen_orbitals)):
            raise TypeError("frozen_orbitals argument must be an (or a list of) integer(s).")

        occupied = [i for i in range(self.n_mos) if self.mo_occ[i] > 0.]
        virtual = [i for i in range(self.n_mos) if self.mo_occ[i] == 0.]

        frozen_occupied = [i for i in frozen_orbitals if i in occupied]
        frozen_virtual = [i for i in frozen_orbitals if i in virtual]

        # Redefined active orbitals based on frozen ones.
        active_occupied = [i for i in occupied if i not in frozen_occupied]
        active_virtual = [i for i in virtual if i not in frozen_virtual]

        # Calculate number of active electrons and active_mos
        n_active_electrons = round(sum([self.mo_occ[i] for i in active_occupied]))
        n_active_mos = len(active_occupied + active_virtual)

        # Exception raised here if there is no active electron.
        # An exception is raised also if all active orbitals are fully occupied.
        if n_active_electrons == 0:
            raise ValueError("There are no active electrons.")
        if n_active_electrons == 2*n_active_mos:
            raise ValueError("All active orbitals are fully occupied.")

        return active_occupied, frozen_occupied, active_virtual, frozen_virtual

    def freeze_mos(self, frozen_orbitals, inplace=True):
        """This method recomputes frozen orbitals with the provided input."""

        list_of_active_frozen = self._convert_frozen_orbitals(frozen_orbitals)

        if any([self.mo_occ[i] == 1 for i in list_of_active_frozen[1]]):
            raise NotImplementedError("Freezing half-filled orbitals is not implemented yet.")

        if inplace:
            self.frozen_orbitals = frozen_orbitals

            self.active_occupied = list_of_active_frozen[0]
            self.frozen_occupied = list_of_active_frozen[1]
            self.active_virtual = list_of_active_frozen[2]
            self.frozen_virtual = list_of_active_frozen[3]

            return None
        else:
            # Shallow copy is enough to copy the same object and changing frozen
            # orbitals (deepcopy also returns errors).
            copy_self = copy.copy(self)

            copy_self.frozen_orbitals = frozen_orbitals

            copy_self.active_occupied = list_of_active_frozen[0]
            copy_self.frozen_occupied = list_of_active_frozen[1]
            copy_self.active_virtual = list_of_active_frozen[2]
            copy_self.frozen_virtual = list_of_active_frozen[3]

            return copy_self

    def energy_from_rdms(self, one_rdm, two_rdm):
        """Computes the molecular energy from one- and two-particle reduced
        density matrices (RDMs). Coefficients (integrals) are computed
        on-the-fly using a pyscf object and the mean-field. Frozen orbitals
        are supported with this method.

        Args:
            one_rdm (numpy.array): One-particle density matrix in MO basis.
            two_rdm (numpy.array): Two-particle density matrix in MO basis.

        Returns:
            float: Molecular energy.
        """

        core_constant, one_electron_integrals, two_electron_integrals = self.get_active_space_integrals()

        # PQRS convention in openfermion:
        # h[p,q]=\int \phi_p(x)* (T + V_{ext}) \phi_q(x) dx
        # h[p,q,r,s]=\int \phi_p(x)* \phi_q(y)* V_{elec-elec} \phi_r(y) \phi_s(x) dxdy
        # The convention is not the same with PySCF integrals. So, a change is
        # reverse back after performing the truncation for frozen orbitals
        two_electron_integrals = two_electron_integrals.transpose(0, 3, 1, 2)

        # Computing the total energy from integrals and provided RDMs.
        e = core_constant + np.sum(one_electron_integrals * one_rdm) + 0.5*np.sum(two_electron_integrals * two_rdm)

        return e.real

    def get_active_space_integrals(self, mo_coeff=None):
        """Computes core constant, one_body, and two-body coefficients with frozen orbitals folded into one-body coefficients
        and core constant

        Args:
            mo_coeff (array): The molecular orbital coefficients to use to generate the integrals

        Returns:
            (float, array, array): (core_constant, one_body coefficients, two_body coefficients)
        """

        return self.get_integrals(mo_coeff, True)

    def get_full_space_integrals(self, mo_coeff=None):
        """Computes core constant, one_body, and two-body integrals for all orbitals

        Args:
            mo_coeff (array): The molecular orbital coefficients to use to generate the integrals.

        Returns:
            (float, array, array): (core_constant, one_body coefficients, two_body coefficients)
        """

        return self.get_integrals(mo_coeff, False)

    def get_integrals(self, mo_coeff=None, consider_frozen=True):
        """Computes core constant, one_body, and two-body coefficients for a given active space and mo_coeff

        Args:
            mo_coeff (array): The molecular orbital coefficients to use to generate the integrals.
            consider_frozen (bool): If True, the frozen orbitals are folded into the one_body and core constant terms.

        Returns:
            (float, array, array): (core_constant, one_body coefficients, two_body coefficients)
        """

        # Pyscf molecule to get integrals.
        pyscf_mol = self.to_pyscf(self.basis, self.symmetry, self.ecp)
        if mo_coeff is None:
            mo_coeff = self.mean_field.mo_coeff

        # Corresponding to nuclear repulsion energy and static coulomb energy.
        core_constant = float(pyscf_mol.energy_nuc())

        # get_hcore is equivalent to int1e_kin + int1e_nuc.
        one_electron_integrals = mo_coeff.T @ self.mean_field.get_hcore() @ mo_coeff

        # Getting 2-body integrals in atomic and converting to molecular basis.
        two_electron_integrals = ao2mo.kernel(pyscf_mol.intor("int2e"), mo_coeff)
        two_electron_integrals = ao2mo.restore(1, two_electron_integrals, len(mo_coeff))

        # PQRS convention in openfermion:
        # h[p,q]=\int \phi_p(x)* (T + V_{ext}) \phi_q(x) dx
        # h[p,q,r,s]=\int \phi_p(x)* \phi_q(y)* V_{elec-elec} \phi_r(y) \phi_s(x) dxdy
        # The convention is not the same with PySCF integrals. So, a change is
        # made before performing the truncation for frozen orbitals.
        two_electron_integrals = two_electron_integrals.transpose(0, 2, 3, 1)
        if consider_frozen:
            core_offset, one_electron_integrals, two_electron_integrals = of_get_active_space_integrals(one_electron_integrals,
                                                                                                        two_electron_integrals,
                                                                                                        self.frozen_occupied,
                                                                                                        self.active_mos)

            # Adding frozen electron contribution to core constant.
            core_constant += core_offset

        return core_constant, one_electron_integrals, two_electron_integrals



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file provides an API enabling the use of randomized classical shadows.
This algorithm is described in H.Y. Huang, R. Kueng, and J. Preskill, Nature
Physics 16, 1050 (2020).
"""

import random

import numpy as np

from tangelo.toolboxes.measurements import ClassicalShadow
from tangelo.linq.circuit import Circuit
from tangelo.linq.helpers.circuits.measurement_basis import measurement_basis_gates, pauli_string_to_of

# State |0> or |1>.
zero_state = np.array([1, 0])
one_state = np.array([0, 1])

# Pauli matrices.
I = np.array([[1, 0], [0, 1]])
X = np.array([[0, 1], [1, 0]])
Y = np.array([[0, -1j], [1j, 0]], dtype=complex)
Z = np.array([[1, 0], [0, -1]])
pauli_matrices = {"X": X, "Y": Y, "Z": Z}

# Traces of each Pauli matrices.
pauli_traces = {pauli: np.trace(matrix) for pauli, matrix in pauli_matrices.items()}

# Reverse channels to undo single Pauli rotations.
S_T = np.array([[1, 0], [0, -1j]], dtype=complex)
H = np.array([[1, 1], [1, -1]]) / np.sqrt(2)
rotations = {"X": H, "Y": H @ S_T, "Z": I}


class RandomizedClassicalShadow(ClassicalShadow):
    r"""Classical shadows using randomized single Pauli measurements, as defined
    in H.Y. Huang, R. Kueng, and J. Preskill, Nature Physics 16, 1050 (2020). In
    short, the channel is inversed to geet the state with the formula
    \hat{\rho} = \bigotimes_{j=1}^n \left( 3U_j^{\dagger} |b_j\rangle \langle b_j| U_j - \mathbb{I} \right)
    """

    def __init__(self, circuit=None, bitstrings=None, unitaries=None, shuffle=True):
        """Overloads the init method to shuffle the bistrings and unitaries if
        those are provided.

        Args:
            shuffle (bool): Randomize bitstrings and unitaries. Default = True.
        """

        super().__init__(circuit, bitstrings, unitaries)

        if bitstrings and shuffle:
            # Shuffling the order while keeping the bistring to its unitary.
            random_bitstrings = list(zip(bitstrings, unitaries))
            random.shuffle(random_bitstrings)
            new_bistrings, new_unitaries = zip(*random_bitstrings)
            self.bitstrings, self.unitaries = list(new_bistrings), list(new_unitaries)

    def build(self, n_shots):
        """Random sampling of single pauli words.

        Args:
            n_shots (int): Total number of measurements.

        Returns:
            list of str: Measurements generated for a randomized procedure.
        """
        measurement_procedure = []
        for _ in range(n_shots):
            single_round_measurement = "".join([random.choice(["X", "Y", "Z"]) for _ in range(self.n_qubits)])
            measurement_procedure.append(single_round_measurement)

        self.unitaries += measurement_procedure
        return measurement_procedure

    def get_basis_circuits(self, only_unique=False):
        """Output a list of circuits corresponding to the random Pauli words
        unitaries.

        Args:
            only_unique (bool): Considering only unique unitaries.

        Returns:
            list of Circuit or tuple: All basis circuits or a tuple of unique
                circuits, string representation and numbers of occurence.
        """

        if not self.unitaries:
            raise ValueError(f"A set of unitaries must de defined (can be done with the build method in {self.__class__.__name__}).")

        unitaries_to_convert = self.unique_unitaries if only_unique else self.unitaries

        basis_circuits = list()
        for pauli_word in unitaries_to_convert:
            # Transformation of a unitary to quantum gates.
            pauli_of = pauli_string_to_of(pauli_word)
            basis_circuits += [Circuit(measurement_basis_gates(pauli_of), self.n_qubits)]

        # Counting each unique circuits (use for reversing to a full shadow from an experiement on hardware).
        if only_unique:
            unique_basis_circuits = [(basis_circuits[i], u, self.unitaries.count(u)) for i, u in enumerate(unitaries_to_convert)]
            return unique_basis_circuits
        else:
            return basis_circuits

    def estimate_state(self, start=0, end=None, indices=None):
        """Returns the classical shadow average density matrix for a range of
        snapshots.

        Args:
            start (int): Starting snapshot for the desired range.
            end (int): Ending snapshot for the desired range.
            indices (list int): Specific snapshot to pick. If this
                variable is set, start and end are ignored.

        Returns:
            array of complex: Estimation of the 2^n * 2^n state.
        """

        # Select specific snapshots. Default: all snapshots.
        if indices is not None:
            snapshot_indices = indices
        else:
            if end is None:
                end = self.size
            snapshot_indices = list(range(start, min(end, self.size)))

        # Creation of the density matrix object that snapshot rho will be added to.
        rho = np.zeros((2**self.n_qubits, 2**self.n_qubits), dtype=complex)

        # Undo rotations for the selected snapshot(s).
        for i_snapshot in snapshot_indices:

            # Starting point is the Identity matrix of size 1.
            rho_snapshot = np.ones((1, 1), dtype=complex)

            for n in range(self.n_qubits):
                state = zero_state if self.bitstrings[i_snapshot][n] == "0" else one_state

                # Unitary to undo the rotation.
                U = rotations[self.unitaries[i_snapshot][n]]
                bU = state @ U
                rho_i = 3 * np.outer(bU.conj(), bU) - I
                rho_snapshot = np.kron(rho_snapshot, rho_i)

            rho += rho_snapshot[:, :]

        # Save the result, return the state estimation
        self.state_estimate = rho / len(snapshot_indices)
        return self.state_estimate

    def get_term_observable(self, term, coeff=1., k=10):
        """Returns the estimated observable for a term and its coefficient.

        Args:
            term (tuple): Openfermion style of a qubit operator term.
            coeff (float): Multiplication factor for the term.
            k (int): Grouping k observations for the means of median protocol.

        Returns:
            float: Observable estimated with the shadow.
        """

        shadow_size = len(self)
        dict_term = dict(term)
        observables_to_median = list()

        shadow_step = shadow_size // k

        # Median of average loop.
        for i_snapshot in range(0, shadow_size - shadow_step, shadow_step):
            observables_to_mean = np.empty(shadow_step, dtype=float)

            for j_snapshot in range(i_snapshot, i_snapshot + shadow_step):
                # Uses the fact that the trace of a tensor product is the
                # product of the traces of the indidual matrices
                # Tr(L \otimes M) = Tr(L) * Tr(M)
                # Also uses the fact that the product of two matrices that
                # are tensor products of smaller matrices is the tensor
                # products of the smaller matrices multiplied
                # (L \otimes M) @ (O \otimes P) = (L @ O) \otimes (M @ P)
                trace = 1.
                for i_qubit in range(self.n_qubits):
                    # If there is an operation applied on the qubit n.
                    if i_qubit in dict_term.keys():
                        obs = pauli_matrices[dict_term[i_qubit]]
                        tobs = pauli_traces[dict_term[i_qubit]]
                    else:
                        # Trace of identity matrix
                        obs = I
                        tobs = 2

                    state = zero_state if self.bitstrings[j_snapshot][i_qubit] == "0" else one_state
                    U = rotations[self.unitaries[j_snapshot][i_qubit]]

                    # Make <b_n|U_n for one qubit.
                    right_side = state @ U

                    # Make obs_n @ U_n^+ @ |b_n> for one qubit.
                    left_side = obs @ right_side.conj()

                    # Below is the faster way to compute
                    # trace *= np.trace(3*np.outer(left_side, right_side) - obs)
                    trace *= 3*np.dot(left_side, right_side) - tobs

                observables_to_mean[j_snapshot - i_snapshot] = np.real(coeff*trace)
            observables_to_median.append(np.mean(observables_to_mean))

        return np.median(observables_to_median)



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import numpy as np
import os
from openfermion import load_operator

from tangelo.molecule_library import mol_H2_sto3g, mol_H4_sto3g, mol_H4_doublecation_minao, mol_H4_cation_sto3g
from tangelo.toolboxes.qubit_mappings import jordan_wigner
from tangelo.toolboxes.ansatz_generator.uccsd import UCCSD
from tangelo.linq import Simulator

# For openfermion.load_operator function.
pwd_this_test = os.path.dirname(os.path.abspath(__file__))


class UCCSDTest(unittest.TestCase):

    def test_uccsd_set_var_params(self):
        """Verify behavior of set_var_params for different inputs (keyword, list, numpy array).
        MP2 have their own tests.
        """

        uccsd_ansatz = UCCSD(mol_H2_sto3g)

        two_ones = np.ones((2,))

        uccsd_ansatz.set_var_params("ones")
        np.testing.assert_array_almost_equal(uccsd_ansatz.var_params, two_ones, decimal=6)

        uccsd_ansatz.set_var_params([1., 1.])
        np.testing.assert_array_almost_equal(uccsd_ansatz.var_params, two_ones, decimal=6)

        uccsd_ansatz.set_var_params(np.array([1., 1.]))
        np.testing.assert_array_almost_equal(uccsd_ansatz.var_params, two_ones, decimal=6)

    def test_uccsd_incorrect_number_var_params(self):
        """Return an error if user provide incorrect number of variational
        parameters.
        """

        uccsd_ansatz = UCCSD(mol_H2_sto3g)

        self.assertRaises(ValueError, uccsd_ansatz.set_var_params, np.array([1., 1., 1., 1.]))

    def test_uccsd_set_params_MP2_H2(self):
        """Verify closed-shell UCCSD functionalities for H2: MP2 initial
        parameters.
        """

        uccsd_ansatz = UCCSD(mol_H2_sto3g)
        uccsd_ansatz.set_var_params("MP2")

        expected = [2e-05, 0.0363253711023451]
        self.assertAlmostEqual(np.linalg.norm(uccsd_ansatz.var_params), np.linalg.norm(expected), delta=1e-10)

    def test_uccsd_set_params_mp2_H2(self):
        """Verify closed-shell UCCSD functionalities for H2: lower case mp2
        initial parameters.
        """

        uccsd_ansatz = UCCSD(mol_H2_sto3g)
        uccsd_ansatz.set_var_params("mp2")

        expected = [2e-05, 0.0363253711023451]
        self.assertAlmostEqual(np.linalg.norm(uccsd_ansatz.var_params), np.linalg.norm(expected), delta=1e-10)

    def test_uccsd_set_params_MP2_H4(self):
        """Verify closed-shell UCCSD functionalities for H4: MP2 initial parameters """

        uccsd_ansatz = UCCSD(mol_H4_sto3g)
        uccsd_ansatz.set_var_params("MP2")

        expected = [2e-05, 2e-05, 2e-05, 2e-05, 0.03894901872789466, 0.07985689676283764, 0.02019977190077326,
                    0.03777151472046017, 0.05845449631119356, 0.017956568628560945, -0.07212522602179856,
                    -0.03958975799697206, -0.042927857009029735, -0.025307140867721886]
        self.assertAlmostEqual(np.linalg.norm(uccsd_ansatz.var_params), np.linalg.norm(expected), delta=1e-10)

    def test_uccsd_H2(self):
        """Verify closed-shell UCCSD functionalities for H2."""

        # Build circuit
        uccsd_ansatz = UCCSD(mol_H2_sto3g)
        uccsd_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = jordan_wigner(mol_H2_sto3g.fermionic_hamiltonian)

        # Assert energy returned is as expected for given parameters
        sim = Simulator()
        uccsd_ansatz.update_var_params([5.86665842e-06, 0.0565317429])
        energy = sim.get_expectation_value(qubit_hamiltonian, uccsd_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.137270174551959, delta=1e-6)

    def test_uccsd_H4_open(self):
        """Verify open-shell UCCSD functionalities for H4."""

        var_params = [-3.68699814e-03,  1.96987010e-02, -1.12573056e-03,  1.31279980e-04,
                      -4.78466616e-02, -8.56400635e-05,  1.60914422e-03, -2.92334744e-02,
                       3.08405067e-03,  1.39404070e-01,  2.42040971e-02, -2.80714763e-03,
                       1.34675820e-01, -7.70867505e-02,  9.86158364e-05,  2.54962567e-02,
                       5.78169071e-02,  2.46873743e-03, -1.05736505e-01, -4.22089003e-02]

        # Build circuit
        uccsd_ansatz = UCCSD(mol_H4_cation_sto3g)
        uccsd_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = jordan_wigner(mol_H4_cation_sto3g.fermionic_hamiltonian)

        # Assert energy returned is as expected for given parameters
        sim = Simulator()
        uccsd_ansatz.update_var_params(var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, uccsd_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.639461490, delta=1e-6)

    def test_uccsd_H4_doublecation(self):
        """Verify closed-shell UCCSD functionalities for H4 2+."""

        var_params = [-1.32047062e-02, -7.16419743e-06, -9.25426159e-06,
                       6.84650642e-02,  6.32462456e-02,  1.44675096e-02,
                      -8.34820283e-06, -7.79703747e-06,  3.28660359e-02]

        # Build circuit
        uccsd_ansatz = UCCSD(mol_H4_doublecation_minao)
        uccsd_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = load_operator("mol_H4_doublecation_minao_qubitham_jw.data", data_directory=pwd_this_test+"/data", plain_text=True)

        # Assert energy returned is as expected for given parameters
        sim = Simulator()
        uccsd_ansatz.update_var_params(var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, uccsd_ansatz.circuit)
        self.assertAlmostEqual(energy, -0.854607, delta=1e-4)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tangelo.helpers.utils import installed_backends

from .translate_braket import translate_braket, get_braket_gates
from .translate_qiskit import translate_qiskit, get_qiskit_gates
from .translate_qulacs import translate_qulacs, get_qulacs_gates
from .translate_cirq import translate_cirq, get_cirq_gates
from .translate_json_ionq import translate_json_ionq, get_ionq_gates
from .translate_qdk import translate_qsharp, get_qdk_gates
from .translate_projectq import translate_projectq, _translate_projectq2abs, get_projectq_gates
from .translate_openqasm import translate_openqasm, _translate_openqasm2abs, get_openqasm_gates


# List all supported gates for all backends found
SUPPORTED_GATES = dict()

SUPPORTED_GATES["projectq"] = sorted(get_projectq_gates().keys())
SUPPORTED_GATES["ionq"] = sorted(get_ionq_gates().keys())
SUPPORTED_GATES["qdk"] = sorted(get_qdk_gates().keys())
SUPPORTED_GATES["openqasm"] = sorted(get_openqasm_gates().keys())

for v in installed_backends:
    SUPPORTED_GATES[v] = sorted(eval(f'get_{v}_gates().keys()'))



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Construct one-particle RDM for DMET calculation.

Construction of the one-particle reduced density matrix (RDM) is done here.
"""

import numpy as np
from functools import reduce


def dmet_low_rdm(active_fock, number_active_electrons):
    """Construct the one-particle RDM from low-level calculation.

    Args:
        active_fock (numpy.array): Fock matrix from low-level calculation
            (float64).
        number_active_electrons (int): Number of electrons in the entire system.

    Returns:
        numpy.array: One-particle RDM of the low-level calculation (float64).
    """

    # Extract the occupied part of the one-particle RDM
    num_occ = number_active_electrons / 2
    e, c = np.linalg.eigh(active_fock)
    new_index = e.argsort()
    e = e[new_index]
    c = c[:, new_index]
    onerdm = np.dot(c[:, : int(num_occ)], c[:, : int(num_occ)].T) * 2

    return onerdm


def dmet_fragment_rdm(t_list, bath_orb, core_occupied, number_active_electrons):
    """Construct the one-particle RDM for the core orbitals.

    Args:
        t_list (list): Number of [0] fragment & [1] bath orbitals (int).
        bath_orb (numpy.array): The bath orbitals (float64).
        core_occupied (numpy.array): Core occupied part of the MO coefficients
            (float64).
        number_active_electrons (int): Number of electrons in the entire system.

    Returns:
        int: Number of orbitals for fragment calculation.
        int: Number of electrons for fragment calulation.
        numpy.array: Core part of the one-particle RDM (float64).
    """

    # Obtain number of active orbitals
    number_orbitals = t_list[0] + t_list[1]

    # Round the values above or below threshold
    for i, core in enumerate(core_occupied):
        if (core < 0.01):
            core_occupied[i] = 0.0
        elif (core > 1.99):
            core_occupied[i] = 2.0

    # Define the number of electrons in the fragment
    number_ele_temp = np.sum(core_occupied)
    number_electrons = int(round(number_active_electrons - number_ele_temp))

    # Obtain the one particle RDM for the fragment (core)
    core_occupied_onerdm = reduce(np.dot, (bath_orb, np.diag(core_occupied), bath_orb.T))

    return number_orbitals, number_electrons, core_occupied_onerdm



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    A test class to check that the circuit class is working properly, and processing the information
    of the gate objects it holds is as expected.
"""

import unittest
import copy
from math import pi
from collections import Counter

from tangelo.linq import Gate, Circuit, stack

# Create several abstract circuits with different features
mygates = [Gate("H", 2), Gate("CNOT", 1, control=0), Gate("CNOT", 2, control=1),
           Gate("Y", 0), Gate("RX", 1, parameter=2.)]

circuit1 = Circuit()
for gate in mygates:
    circuit1.add_gate(gate)
circuit2 = Circuit(mygates)

circuit3 = Circuit(mygates)
circuit3.add_gate(Gate("RZ", 4, parameter="some angle", is_variational=True))

circuit4 = copy.deepcopy(circuit3)
circuit4.add_gate(Gate("RY", 3, parameter="some angle", is_variational=True))

entangle_circuit = Circuit([Gate("CSWAP", target=[2, 5], control=[0]),
                            Gate("CSWAP", target=[3, 7], control=[4]),
                            Gate("H", 6)], n_qubits=10)


class TestCircuits(unittest.TestCase):

    def test_init(self):
        """ Tests that constructing a circuit consisting of several instructions works through either passing
        a list of gates directly , or by using the add_gate method repeatedly on an existing circuit. Simply test
        that the number of gates is what was expected """

        self.assertTrue(circuit1.size == len(mygates))
        self.assertTrue(circuit2.size == len(mygates))

    def test_is_variational(self):
        """ Ensure that the circuit is labeled as variational as soon as one variational gate is present """

        self.assertTrue(circuit1.is_variational is False)
        self.assertTrue(circuit3.is_variational is True)

    def test_gate_data_is_copied(self):
        """ Ensure that circuit is not referencing mutable variables that could cause it to change after
        instantiation if the values of the variables are later changed in external code. """

        mygates2 = copy.deepcopy(mygates)
        c1 = Circuit(mygates2)

        g = mygates2[0]
        g.target.append(1)
        g.name = 'POTATO'
        g.parameter = -999.

        c2 = Circuit(mygates2)
        self.assertTrue(c1 != c2)

    def test_width(self):
        """ Ensure the width attribute of the circuit object (number of qubits) matches the gate operations
        present in the circuit. """

        self.assertTrue(circuit1.width == 3)
        self.assertTrue(circuit3.width == 5)

    def test_gate_counts(self):
        """ Test that all gates have been counted """

        n_gates = sum(circuit1._gate_counts.values())
        self.assertTrue(circuit1.size == n_gates)

    def test_add_circuits(self):
        """ Test the concatenation of two circuit objects """

        circuit_sum = circuit3 + circuit4
        self.assertTrue(circuit_sum.size == circuit3.size + circuit4.size)
        self.assertTrue(circuit_sum.is_variational == circuit3.is_variational or circuit4.is_variational)
        self.assertTrue(circuit_sum._qubit_indices == circuit3._qubit_indices.union(circuit4._qubit_indices))
        self.assertTrue(circuit_sum._gate_counts == dict(Counter(circuit3._gate_counts)
                                                         + Counter(circuit4._gate_counts)))
        self.assertTrue(len(circuit_sum._variational_gates) == (len(circuit3._variational_gates) +
                                                                len(circuit4._variational_gates)))

    def test_mul_circuit(self):
        """ Test the multiplication (repetition) operator for circuit objects """

        # Should work for *2
        c2 = circuit3 * 2
        ref_counts = {k: 2*v for k, v in circuit3._gate_counts.items()}
        self.assertTrue(c2._gate_counts == ref_counts)
        self.assertTrue(len(c2._variational_gates) == 2*len(circuit3._variational_gates))

        # Fail for incorrect values, such as 2.5, or 0.
        with self.assertRaises(ValueError):
            _ = circuit3 * 2.5
        with self.assertRaises(ValueError):
            _ = circuit3 * 0

        # Repeating an empty circuit yields an empty circuit
        self.assertTrue((Circuit()*3).size == 0)

        # Check on right-hand side
        self.assertTrue(2*circuit3 == c2)

    def test_entangled_indices(self):
        """ Test that entangled indices subsets are properly updated after
        a new gate is added to the circuit. """

        c = Circuit()
        c.add_gate(Gate("CNOT", target=4, control=0))
        self.assertTrue(c.get_entangled_indices() == [{0, 4}])
        c.add_gate(Gate("CNOT", target=5, control=1))
        self.assertTrue(c.get_entangled_indices() == [{0, 4}, {1, 5}])
        c.add_gate(Gate("H", target=2))
        self.assertTrue(c.get_entangled_indices() == [{0, 4}, {1, 5}, {2}])
        c.add_gate(Gate("CNOT", target=6, control=5))
        self.assertTrue(c.get_entangled_indices() == [{0, 4}, {2}, {1, 5, 6}])
        c.add_gate(Gate("CNOT", target=1, control=4))
        self.assertTrue(c.get_entangled_indices() == [{2}, {0, 1, 4, 5, 6}])
        c.add_gate(Gate("CSWAP", target=[2, 7], control=[0]))
        self.assertTrue(c.get_entangled_indices() == [{0, 1, 2, 4, 5, 6, 7}])

    def test_trim_circuit(self):
        """ Check that unnecessary indices are trimmed and new indices minimal """

        ref_c = Circuit([Gate("CSWAP", target=[1, 4], control=[0]),
                         Gate("CSWAP", target=[2, 6], control=[3]),
                         Gate("H", 5)], n_qubits=7)
        entangle_circuit.trim_qubits()
        self.assertTrue(ref_c == entangle_circuit)

    def test_reindex_qubits(self):
        """ Test the function that reindexes qubits (e.g replaces indices by another). """

        # With circuit of natural width
        gates = [Gate("H", 2), Gate("CNOT", 1, control=0), Gate("CSWAP", target=[1, 2], control=[0])]
        c1 = Circuit(copy.deepcopy(gates))
        c1.reindex_qubits([4, 5, 6])

        ref = [Gate("H", 6), Gate("CNOT", 5, control=4), Gate("CSWAP", target=[5, 6], control=[4])]
        self.assertTrue(ref == c1._gates)

        # With circuit of fixed width (sends 4,5,6 to 0,1,2, the rest is not relevant)
        c2 = Circuit(ref, n_qubits=8)
        c2.reindex_qubits([3, 4, 5, 6, 0, 1, 2, 7])
        self.assertTrue(gates == c2._gates)

        # Test for input of incorrect length in both previous cases
        with self.assertRaises(ValueError):
            c1.reindex_qubits([2])
        with self.assertRaises(ValueError):
            c2.reindex_qubits([0, 1, 2])

    def test_split_circuit(self):
        """ Test function that splits circuit into several circuits targeting qubit subsets
        that are not entangled with each other. Trims unnecessary qubit indices. """
        c = Circuit([Gate("CSWAP", target=[2, 5], control=[0]),
                     Gate("CSWAP", target=[3, 7], control=[4]),
                     Gate("H", 6)])
        c1, c2, c3 = c.split()

        self.assertTrue(c1 == Circuit([Gate("CSWAP", target=[1, 2], control=[0])]))
        self.assertTrue(c2 == Circuit([Gate("CSWAP", target=[0, 2], control=[1])]))
        self.assertTrue(c3 == Circuit([Gate("H", target=0)]))

    def test_stack_circuits(self):
        """ Test circuit stacking """

        c1 = Circuit([Gate("H", 6)])
        c2 = Circuit([Gate("CNOT", 5, control=4)])
        c3 = Circuit([Gate("CSWAP", target=[5, 6], control=[4])])

        ref = [Gate("H", 0), Gate("CNOT", 2, control=1), Gate("CSWAP", target=[4, 5], control=[3])]

        # No and multiple arguments, natural or as an unpacked list
        self.assertTrue(ref == stack(c1, c2, c3)._gates)
        self.assertTrue(ref == stack(*[c1, c2, c3])._gates)
        self.assertTrue([] == stack(*[])._gates)

        # Try convenience method in Circuit class
        self.assertTrue(ref == c1.stack(c2, c3)._gates)

        c4 = Circuit([Gate("H", 0), Gate("CNOT", 1, control=0), Gate("X", 0), Gate("RX", 1, parameter=2.)])

        ref2 = [Gate("H", 0), Gate("CNOT", 1, control=0), Gate("X", 0), Gate("RX", 1, parameter=2.),
                Gate("H", 2), Gate("CNOT", 3, control=2), Gate("X", 2), Gate("RX", 3, parameter=2.)]
        # Stacked copies of same circuit
        self.assertTrue(ref2 == stack(c4, c4)._gates)
        self.assertTrue(ref2 == c4.stack(c4)._gates)

    def test_equality_circuit(self):
        """ Test equality operators (== and !=) for circuits """
        self.assertTrue(circuit1 == circuit2)
        self.assertTrue(circuit3 != circuit2)
        c3 = Circuit(circuit3._gates, n_qubits=6)
        self.assertTrue(circuit3 != c3)

    def test_fixed_sized_circuit_above(self):
        """ If circuit is instantiated with fixed width, the code must throw if qubit indices are not consistent """
        circuit_fixed = Circuit(n_qubits=2)
        gate1 = Gate("H", 2)
        gate2 = Gate("CNOT", 0, control=5)
        self.assertRaises(ValueError, circuit_fixed.add_gate, gate1)
        self.assertRaises(ValueError, circuit_fixed.add_gate, gate2)

    def test_fixed_sized_circuit_below(self):
        """ If circuit is instantiated with fixed width, then the width property must be consistent, regardless
        of what gate instructions have been passed """
        n_qubits = 3
        circuit_fixed = Circuit([Gate("H", 0)], n_qubits=n_qubits)
        self.assertTrue(circuit_fixed.width == n_qubits)

    def test_inverse(self):
        """ Test if inverse function returns the proper set of gates."""

        mygates_inverse = list()
        mygates_inverse.append(Gate("RX", 1, parameter=-2.))
        mygates_inverse.append(Gate("Y", 0))
        mygates_inverse.append(Gate("CNOT", 2, control=1))
        mygates_inverse.append(Gate("CNOT", 1, control=0))
        mygates_inverse.append(Gate("H", 2))
        circuit1_inverse = Circuit(mygates_inverse)
        self.assertTrue(circuit1.inverse(), circuit1_inverse)

        ts_circuit = Circuit([Gate("T", 0), Gate("S", 1)])
        ts_circuit_inverse = Circuit([Gate("PHASE", 0, parameter=-pi/4), Gate("PHASE", 0, parameter=-pi/2)])
        self.assertTrue(ts_circuit.inverse(), ts_circuit_inverse)

    def test_depth(self):
        """ Test depth method on a few circuits """

        c1 = Circuit([Gate("H", 0)]*3 + [Gate("X", 1)])
        self.assertTrue(c1.depth() == 3)

        c2 = Circuit([Gate("H", 0), Gate("CNOT", 1, 0), Gate("CNOT", 2, 1), Gate("H", 0), Gate("CNOT", 0, 2)])
        self.assertTrue(c2.depth() == 4)

        c3 = Circuit()
        self.assertTrue(c3.depth() == 0)

    def test_simple_optimization_functions(self):
        """ Test if removing small rotations and redundant gates return the
        proper set of gates.
        """

        test_circuit = Circuit([Gate("RX", 0, parameter=2.), Gate("CNOT", 1, control=0),
                    Gate("RZ", 2, parameter=0.01), Gate("CNOT", 1, control=0),
                    Gate("RX", 0, parameter=-2.)])
        test_circuit.remove_small_rotations(param_threshold=0.05)

        ref_gates = [Gate("RX", 0, parameter=2.), Gate("CNOT", 1, control=0),
                     Gate("CNOT", 1, control=0), Gate("RX", 0, parameter=-2.)]

        self.assertTrue(ref_gates == test_circuit._gates)

        test_circuit.remove_redundant_gates()

        self.assertTrue([] == test_circuit._gates)

    def test_simple_optimization_minus_a_qubit(self):
        """ Test if removing redundant gates deletes a qubit."""

        test_circuit = Circuit([Gate("X", 0), Gate("H", 1), Gate("H", 1)])
        test_circuit.remove_redundant_gates()

        self.assertEqual(test_circuit.width, 1)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from tangelo.linq import Gate, Circuit
from tangelo.toolboxes.measurements import DerandomizedClassicalShadow
from tangelo.toolboxes.operators import QubitOperator

# Circuit to sample (Bell state).
state = Circuit([Gate("H", 0), Gate("CNOT", 1, 0)])

# Simple saved bistrings and unitaries to construct a shadow (size = 100).
bitstrings = ["11", "01", "11", "01", "00", "10", "00", "10", "11", "01", "00",
              "01", "00", "01", "11", "10", "00", "01", "11", "01", "00", "01",
              "11", "01", "11", "01", "00", "01", "11", "10", "11", "10", "11",
              "10", "00", "10", "00", "10", "00", "01", "11", "10", "11", "10",
              "00", "01", "11", "01", "11", "10", "00", "10", "00", "01", "11",
              "10", "00", "10", "11", "10", "11", "10", "00", "01", "11", "01",
              "11", "01", "11", "10", "00", "01", "11", "01", "11", "10", "11",
              "01", "11", "10", "11", "10", "00", "10", "11", "01", "00", "10",
              "00", "10", "11", "10", "11", "01", "11", "01", "00", "10", "11",
              "10"]
unitaries = ["XX", "YY"] * 50


class DerandomizedClassicalShadowTest(unittest.TestCase):

    def test_initialization(self):
        """Testing the initialization."""

        DerandomizedClassicalShadow(state, bitstrings, unitaries)

    def test_same_length_check(self):
        """Testing the case where arguments are not the same length."""

        wrong_bitstrings = bitstrings + ["00"]
        with self.assertRaises(AssertionError):
            DerandomizedClassicalShadow(state, wrong_bitstrings, unitaries)

    def test_shadow_properties(self):
        """Testing of the shadow properties."""

        cs = DerandomizedClassicalShadow(state, bitstrings, unitaries)

        self.assertEqual(cs.n_qubits, 2)
        self.assertEqual(cs.size, 100)
        self.assertEqual(len(cs), 100)

    def test_get_term_observable(self):
        """Testing the computation of a single qubit term."""

        cs = DerandomizedClassicalShadow(state, bitstrings, unitaries)

        obs_xx = cs.get_term_observable([(0, "X"), (1, "X")], 1.)
        self.assertAlmostEqual(obs_xx, 1.0, places=4)

        obs_yy = cs.get_term_observable([(0, "Y"), (1, "Y")], 1.)
        self.assertAlmostEqual(obs_yy, -1.0, places=4)

    def test_get_observable(self):
        """Testings the computation of an eigenvalue of a QubitOperator."""

        cs = DerandomizedClassicalShadow(state, bitstrings, unitaries)
        obs_xx = cs.get_observable(QubitOperator("X0 X1", coefficient=1.))
        self.assertAlmostEqual(obs_xx, 1.0, places=4)

        obs_yy = cs.get_observable(QubitOperator("Y0 Y1", coefficient=1.))
        self.assertAlmostEqual(obs_yy, -1.0, places=4)

    def test_get_basis_circuits(self):
        """Testing of the method to get the appended circuit corresponding to
        the unitaries.
        """
        cs = DerandomizedClassicalShadow(state, bitstrings, unitaries)

        self.assertEqual(len(cs.get_basis_circuits(False)), 100)
        self.assertEqual(len(cs.get_basis_circuits(True)), 2)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc


class ProblemDecomposition(abc.ABC):

    def __init__(self):
        self.electronic_structure_solver = None
        self.electron_localization_method = None

    @abc.abstractmethod
    def simulate(self, molecule, fragment_atoms, mean_field=None):
        pass



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Perform fragment SCF calculation.

The fragment SCF calculation for DMET calculation is done here.
"""

from pyscf import gto, scf, ao2mo
from functools import reduce
import numpy as np


def dmet_fragment_scf(t_list, two_ele, fock, number_electrons, number_orbitals, guess_orbitals, chemical_potential):
    """Perform SCF calculation.

    Args:
        t_list (list): Number of [0] fragment & [1] bath orbitals (int).
        two_ele (numpy.array): Two-electron integrals for fragment calculation
            (float64).
        fock (numpy.array): The fock matrix for fragment calculation (float64).
        number_electrons (int): Number of electrons for fragment calculation.
        number_orbitals (int): Number of orbitals for fragment calculation.
        guess_orbitals (numpy.array): Guess orbitals for SCF calculation
            (float64).
        chemical_potential (float64): The chemical potential.

    Returns:
        pyscf.scf.RHF: The mean field of the molecule (Fragment calculation).
        numpy.array: The fock matrix with chemical potential subtracted
            (float64).
        pyscf.gto.Mole: The molecule to simulate (Fragment calculation).
    """

    # Deep copy the fock matrix
    fock_frag_copy = fock.copy()

    # Subtract the chemical potential to make the number of electrons consistent
    if (chemical_potential != 0.0):
        for orb in range(t_list[0]):
            fock_frag_copy[orb, orb] -= chemical_potential

    # Determine the molecular space (set molecule object of pyscf)
    mol_frag = gto.Mole()
    mol_frag.build(verbose=0)
    mol_frag.atom.append(("C", (0, 0, 0)))
    mol_frag.nelectron = number_electrons
    mol_frag.incore_anyway = True

    # Perform SCF calculation (set mean field object of pyscf)
    mf_frag = scf.RHF(mol_frag)
    mf_frag.get_hcore = lambda *args: fock_frag_copy
    mf_frag.get_ovlp = lambda *args: np.eye(number_orbitals)
    mf_frag._eri = ao2mo.restore(8, two_ele, number_orbitals)
    mf_frag.scf(guess_orbitals)

    # Calculate the density matrix for the fragment
    dm_frag = reduce(np.dot, (mf_frag.mo_coeff, np.diag(mf_frag.mo_occ), mf_frag.mo_coeff.T))

    # Use newton-raphson algorithm if the above SCF calculation is not converged
    if (mf_frag.converged is False):
        mf_frag.get_hcore = lambda *args: fock_frag_copy
        mf_frag.get_ovlp = lambda *args: np.eye(number_orbitals)
        mf_frag._eri = ao2mo.restore(8, two_ele, number_orbitals)
        mf_frag = scf.RHF(mol_frag).newton()
        energy = mf_frag.kernel(dm0=dm_frag)
        dm_frag = reduce(np.dot, (mf_frag.mo_coeff, np.diag(mf_frag.mo_occ), mf_frag.mo_coeff.T))

    return mf_frag, fock_frag_copy, mol_frag



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for closed-shell and restricted open-shell qubit coupled cluster (QCC) ansatze. """

import os
import unittest

import numpy as np
from openfermion import load_operator

from tangelo.linq import Simulator
from tangelo.toolboxes.ansatz_generator.qcc import QCC
from tangelo.toolboxes.operators.operators import QubitOperator
from tangelo.molecule_library import mol_H2_sto3g, mol_H4_cation_sto3g, mol_H4_doublecation_minao

sim = Simulator()

# For openfermion.load_operator function.
pwd_this_test = os.path.dirname(os.path.abspath(__file__))


class QCCTest(unittest.TestCase):
    """Unit tests for various functionalities of the QCC ansatz class. Examples for both closed-
    and restricted open-shell QCC are provided using H2, H4+, and H4+2."""

    @staticmethod
    def test_qcc_set_var_params():
        """ Verify behavior of set_var_params for different inputs (keyword, list, numpy array). """

        qcc_ansatz = QCC(mol_H2_sto3g, up_then_down=True)

        nine_zeros = np.zeros((9,), dtype=float)

        qcc_ansatz.set_var_params([0.] * 9)
        np.testing.assert_array_almost_equal(qcc_ansatz.var_params, nine_zeros, decimal=6)

        nine_tenths = 0.1 * np.ones((9,))

        qcc_ansatz.set_var_params([0.1] * 9)
        np.testing.assert_array_almost_equal(qcc_ansatz.var_params, nine_tenths, decimal=6)

        qcc_ansatz.set_var_params(np.array([0.1] * 9))
        np.testing.assert_array_almost_equal(qcc_ansatz.var_params, nine_tenths, decimal=6)

    def test_qcc_incorrect_number_var_params(self):
        """ Return an error if user provide incorrect number of variational parameters """

        qcc_ansatz = QCC(mol_H2_sto3g, up_then_down=True)

        self.assertRaises(ValueError, qcc_ansatz.set_var_params, np.array([1.] * 2))

    def test_qcc_h2(self):
        """ Verify closed-shell functionality when using the QCC class for H2 """

        # Specify the qubit operators from the direct interaction set (DIS) of QCC generators.
        dis = [QubitOperator("Y0 X1 X2 X3")]
        qcc_ansatz = QCC(mol_H2_sto3g, up_then_down=True, dis=dis)

        # Build the QCC circuit, which is prepended by the qubit mean field (QMF) circuit.
        qcc_ansatz.build_circuit()

        # Get qubit hamiltonian for energy evaluation
        qubit_hamiltonian = qcc_ansatz.qubit_ham

        # The QMF and QCC parameters can both be specified; determined automatically othersise.
        qmf_var_params = [ 3.14159265e+00, -2.42743256e-08,  3.14159266e+00, -3.27162543e-08,
                           3.08514545e-09,  3.08514545e-09,  3.08514545e-09,  3.08514545e-09]
        qcc_var_params = [-2.26136280e-01]
        var_params = qmf_var_params + qcc_var_params
        # Assert energy returned is as expected for given parameters
        qcc_ansatz.update_var_params(var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, qcc_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.1372701746609022, delta=1e-6)

    def test_qmf_qcc_h4_cation(self):
        """ Verify restricted open-shell functionality when using QCC ansatz for H4+ """

        # Specify the qubit operators from the direct interaction set (DIS) of QCC generators.
        dis = [QubitOperator("X0 X1 Y2 Y3 X4 Y5"), QubitOperator("Y1 Y3 Y4 X5"),
               QubitOperator("X0 Y1 Y3 Y4"), QubitOperator("X1 X2 Y3 X4 X5"),
               QubitOperator("Y1 Y2 Y3 X4"), QubitOperator("Y1 X3 X4"),
               QubitOperator("Y0 X2"), QubitOperator("X0 X1 X3 X4 Y5"),
               QubitOperator("X0 X1 X2 Y3 X4")]
        qcc_ansatz = QCC(mol_H4_cation_sto3g, "scbk", True, dis)

        # Build the QCC circuit, which is prepended by the qubit mean field (QMF) circuit.
        qcc_ansatz.build_circuit()

        # Get qubit hamiltonian for energy evaluation
        qubit_hamiltonian = qcc_ansatz.qubit_ham

        # The QMF and QCC parameters can both be specified; determined automatically othersise.
        qmf_var_params = [3.14159302e+00,  6.20193478e-07,  1.51226426e-06,  3.14159350e+00,
                          3.14159349e+00,  7.88310582e-07,  3.96032530e+00,  2.26734374e+00,
                          3.22127001e+00,  5.77997401e-01,  5.51422406e+00,  6.26513711e+00]
        qcc_var_params = [-0.26816042,  0.21694796,  0.12139543, -0.2293093,
                          -0.14577423, -0.08937818,  0.01796464, -0.06445363,  0.06056016]
        var_params = qmf_var_params + qcc_var_params
        # Assert energy returned is as expected for given parameters
        qcc_ansatz.update_var_params(var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, qcc_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.6382913, delta=1e-6)

    def test_qmf_qcc_h4_double_cation(self):
        """ Verify restricted open-shell functionality when using the QCC ansatz for H4+2 """

        # Specify the qubit operators from the direct interaction set (DIS) of QCC generators.
        dis = [QubitOperator("Y0 X4"), QubitOperator("X0 Y1 Y2 X4 Y5 X6"),
               QubitOperator("Y0 Y1 Y4 X5"), QubitOperator("Y0 X1 Y4 Y5 X6"),
               QubitOperator("X0 X1 X2 Y4 X5")]
        qcc_ansatz = QCC(mol_H4_doublecation_minao, "bk", True, dis)

        # Build the QCC circuit, which is prepended by the qubit mean field (QMF) circuit.
        qcc_ansatz.build_circuit()

        # Get qubit hamiltonian for energy evaluation
        qubit_hamiltonian = load_operator("mol_H4_doublecation_minao_qubitham_bk_updown.data", data_directory=pwd_this_test+"/data", plain_text=True)

        # The QMF and QCC parameters can both be specified; determined automatically othersise.
        qmf_var_params = [3.14159247e+00,  3.14158884e+00,  1.37660700e-06,  3.14159264e+00,
                          3.14159219e+00,  3.14158908e+00,  0.00000000e+00,  0.00000000e+00,
                          6.94108155e-01,  1.03928030e-01,  5.14029803e+00,  2.81850365e+00,
                          4.25403875e+00,  6.19640367e+00,  1.43241026e+00,  3.50279759e+00]
        qcc_var_params = [-2.76489925e-01, -2.52783324e-01,  5.76565629e-02,  6.99988237e-02,
                          -7.03721438e-02]
        var_params = qmf_var_params + qcc_var_params
        # Assert energy returned is as expected for given parameters
        qcc_ansatz.update_var_params(var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, qcc_ansatz.circuit)
        self.assertAlmostEqual(energy, -0.85465810, delta=1e-6)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions helping with quantum circuit format conversion between abstract
format and Braket format

In order to produce an equivalent circuit for the target backend, it is necessary
to account for:
- how the gate names differ between the source backend to the target backend.
- how the order and conventions for some of the inputs to the gate operations
    may also differ.
"""


def get_braket_gates():
    """Map gate name of the abstract format to the equivalent methods of the
    braket.circuits.Circuit class API and supported gates:
    https://amazon-braket-sdk-python.readthedocs.io/en/latest/_apidoc/braket.circuits.circuit.html
    """

    from braket.circuits import Circuit as BraketCircuit

    GATE_BRAKET = dict()
    GATE_BRAKET["H"] = BraketCircuit.h
    GATE_BRAKET["X"] = BraketCircuit.x
    GATE_BRAKET["Y"] = BraketCircuit.y
    GATE_BRAKET["Z"] = BraketCircuit.z
    GATE_BRAKET["CX"] = BraketCircuit.cnot
    GATE_BRAKET["CY"] = BraketCircuit.cy
    GATE_BRAKET["CZ"] = BraketCircuit.cz
    GATE_BRAKET["S"] = BraketCircuit.s
    GATE_BRAKET["T"] = BraketCircuit.t
    GATE_BRAKET["RX"] = BraketCircuit.rx
    GATE_BRAKET["RY"] = BraketCircuit.ry
    GATE_BRAKET["RZ"] = BraketCircuit.rz
    GATE_BRAKET["XX"] = BraketCircuit.xx
    GATE_BRAKET["CRZ"] = [BraketCircuit.cphaseshift, BraketCircuit.cphaseshift10]
    GATE_BRAKET["PHASE"] = BraketCircuit.phaseshift
    GATE_BRAKET["CPHASE"] = BraketCircuit.cphaseshift
    GATE_BRAKET["CNOT"] = BraketCircuit.cnot
    GATE_BRAKET["SWAP"] = BraketCircuit.swap
    GATE_BRAKET["CSWAP"] = BraketCircuit.cswap
    # GATE_BRAKET["MEASURE"] = ? (mid-circuit measurement currently unsupported?)

    return GATE_BRAKET


def translate_braket(source_circuit):
    """Take in an abstract circuit, return a quantum circuit object as defined
    in the Python Braket SDK.

    Args:
        source_circuit: quantum circuit in the abstract format.

    Returns:
        braket.circuits.Circuit: quantum circuit in Python Braket SDK format.
    """

    from braket.circuits import Circuit as BraketCircuit

    GATE_BRAKET = get_braket_gates()
    target_circuit = BraketCircuit()

    # Map the gate information properly. Different for each backend (order, values)
    for gate in source_circuit._gates:
        if gate.control is not None:
            if len(gate.control) > 1:
                raise ValueError('Multi-controlled gates not supported with braket: Gate {gate.name} with controls {gate.control} is invalid')
        if gate.name in {"H", "X", "Y", "Z", "S", "T"}:
            (GATE_BRAKET[gate.name])(target_circuit, gate.target[0])
        elif gate.name in {"RX", "RY", "RZ", "PHASE"}:
            (GATE_BRAKET[gate.name])(target_circuit, gate.target[0], gate.parameter)
        elif gate.name in {"CNOT", "CX", "CY", "CZ"}:
            (GATE_BRAKET[gate.name])(target_circuit, control=gate.control[0], target=gate.target[0])
        elif gate.name in {"XX"}:
            (GATE_BRAKET[gate.name])(target_circuit, gate.target[0], gate.target[1], gate.parameter)
        elif gate.name in {"CRZ"}:
            (GATE_BRAKET[gate.name][0])(target_circuit, gate.control[0], gate.target[0], gate.parameter/2.)
            (GATE_BRAKET[gate.name][1])(target_circuit, gate.control[0], gate.target[0], -gate.parameter/2.)
        elif gate.name in {"SWAP"}:
            (GATE_BRAKET[gate.name])(target_circuit, gate.target[0], gate.target[1])
        elif gate.name in {"CSWAP"}:
            (GATE_BRAKET[gate.name])(target_circuit, gate.control[0], gate.target[0], gate.target[1])
        elif gate.name in {"CPHASE"}:
            (GATE_BRAKET[gate.name])(target_circuit, gate.control[0], gate.target[0], gate.parameter)
        # elif gate.name in {"MEASURE"}:
        # implement if mid-circuit measurement available through Braket later on
        else:
            raise ValueError(f"Gate '{gate.name}' not supported on backend braket")
    return target_circuit



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Perform IAO localization.

The orbital localization of the canonical orbitals using Intrinsic Atomic
Orbitals (IAO) localization is done here. `pyscf.lo` is used.

Note that minimal basis cannot be used for IAO because the idea of IAO is to map
on minao minimal basis set.

For details, refer to:
    - G. Knizia, JCTC 9, 4834-4843 (2013).
"""

from pyscf import gto
from pyscf.lo import iao
from functools import reduce
from pyscf.lo import orth
import numpy as np
import scipy


def iao_localization(mol, mf):
    """Localize the orbitals using IAO localization.

    Args:
        mol (pyscf.gto.Mole): The molecule to simulate.
        mf (pyscf.scf.RHF): The mean field of the molecule.

    Returns:
        numpy.array: The localized orbitals (float64).
    """

    if mol.basis in {"minao", "sto-3g", "sto-6g"}:
        raise RuntimeError("Using IAO localization with minimal basis is not supported.")

    #   Construct IAO from occupied orbitals
    iao1 = _iao_occupied_orbitals(mol, mf)

    #   Construct IAO from complementary space
    iao2 = _iao_complementary_orbitals(mol, iao1)

    #   Gather two and assign the IAOs to atoms, rearrange them
    iao_lo = _iao_atoms(mol, iao1, iao2)

    return iao_lo


def _iao_occupied_orbitals(mol, mf):
    """Get the IAOs for occupied space.

    Args:
        mol (pyscf.gto.Mole): The molecule to simulate.
        mf (pyscf.scf.RHF): The mean field of the molecule.

    Returns:
        numpy.array: The localized orbitals for the occupied space (float64).
    """

    #   Get MO coefficient of occupied MOs
    occupied_orbitals = mf.mo_coeff[:, mf.mo_occ > 0.5]

    #   Get mol data in minao basis
    min_mol = iao.reference_mol(mol)

    #   Calculate the overlaps for total basis
    s1 = mol.intor_symmetric("int1e_ovlp")

    #   ... for minao basis
    s2 = min_mol.intor_symmetric("int1e_ovlp")

    #   ... between the two basis (and transpose)
    s12 = gto.mole.intor_cross("int1e_ovlp", mol, min_mol)
    s21 = s12.T

    #   Calculate P_12 = S_1^-1 * S_12 using Cholesky decomposition
    s1_sqrt = scipy.linalg.cho_factor(s1)
    s2_sqrt = scipy.linalg.cho_factor(s2)
    p12 = scipy.linalg.cho_solve(s1_sqrt, s12)

    #   C~ = second_half ( S_1^-1 * S_12 * first_half ( S_2^-1 * S_21 * C ) )
    c_tilde = scipy.linalg.cho_solve(s2_sqrt, np.dot(s21, occupied_orbitals))
    c_tilde = scipy.linalg.cho_solve(s1_sqrt, np.dot(s12, c_tilde))
    c_tilde = np.dot(c_tilde, orth.lowdin(reduce(np.dot, (c_tilde.T, s1, c_tilde))))

    #   Obtain C * C^T * S1 and C~ * C~^T * S1
    ccs1 = reduce(np.dot, (occupied_orbitals, occupied_orbitals.conj().T, s1))
    ctcts1 = reduce(np.dot, (c_tilde, c_tilde.conj().T, s1))

    #   Calculate A = ccs1 * ctcts1 * p12 + ( 1 - ccs1 ) * ( 1 - ctcts1 ) * p12
    iao_active = (p12 + reduce(np.dot, (ccs1, ctcts1, p12)) * 2 - np.dot(ccs1, p12) - np.dot(ctcts1, p12))

    #   Orthogonalize A
    iao_active = np.dot(iao_active, orth.lowdin(reduce(np.dot, (iao_active.T, s1, iao_active))))

    return iao_active


def _iao_complementary_orbitals(mol, iao_ref):
    """Get the IAOs for complementary space (virtual orbitals).

    Args:
        mol (pyscf.gto.Mole): The molecule to simulate.
        iao_ref (numpy.array): IAO in occupied space (float64).

    Returns:
        numpy.array: IAO in complementary space (float64).
    """

    #   Get the total number of AOs
    norbital_total = mol.nao_nr()

    #   Calculate the Overlaps for total basis
    s1 = mol.intor_symmetric("int1e_ovlp")

    #   Construct the complementary space AO
    number_iaos = iao_ref.shape[1]
    number_inactive = norbital_total - number_iaos
    iao_com_ref = _iao_complementary_space(iao_ref, s1, number_inactive)

    #   Get a list of active orbitals
    min_mol = iao.reference_mol(mol)
    norbital_active, active_list = _iao_count_active(mol, min_mol)

    #   Obtain the Overlap-like matrices
    s21 = s1[active_list, :]
    s2 = s21[:, active_list]
    s12 = s21.T

    #   Calculate P_12 = S_1^-1 * S_12 using Cholesky decomposition
    s1_sqrt = scipy.linalg.cho_factor(s1)
    s2_sqrt = scipy.linalg.cho_factor(s2)
    p12 = scipy.linalg.cho_solve(s1_sqrt, s12)

    #   C~ = orth ( second_half ( S_1^-1 * S_12 * first_half ( S_2^-1 * S_21 * C ) ) )
    c_tilde = scipy.linalg.cho_solve(s2_sqrt, np.dot(s21, iao_com_ref))
    c_tilde = scipy.linalg.cho_solve(s1_sqrt, np.dot(s12, c_tilde))
    c_tilde = np.dot(c_tilde, orth.lowdin(reduce(np.dot, (c_tilde.T, s1, c_tilde))))

    #   Obtain C * C^T * S1 and C~ * C~^T * S1
    ccs1 = reduce(np.dot, (iao_com_ref, iao_com_ref.conj().T, s1))
    ctcts1 = reduce(np.dot, (c_tilde, c_tilde.conj().T, s1))

    #   Calculate A = ccs1 * ctcts1 * p12 + ( 1 - ccs1 ) * ( 1 - ctcts1 ) * p12
    iao_comp = (p12 + reduce(np.dot, (ccs1, ctcts1, p12)) * 2 - np.dot(ccs1, p12) - np.dot(ctcts1, p12))
    iao_comp = np.dot(iao_comp, orth.lowdin(reduce(np.dot, (iao_comp.T, s1, iao_comp))))

    return iao_comp


def _iao_count_active(mol, min_mol):
    """Figure out the basis functions matching with MINAO.

    Args:
        mol (pyscf.gto.Mole): The molecule to simulate.
        min_mol (numpy.array): The molecule to simulate in minao basis.

    Returns:
        int: Number of active orbitals.
        list: List of active orbitals (int).
    """

    #   Initialize the list
    active_number_list = []

    #   Loop over all basis and see if there are labels matching with the MINAO ones
    for idx, total_basis in enumerate(mol.spheric_labels()):
        if all([min_basis != total_basis for min_basis in min_mol.spheric_labels()]):
            active_number_list.append(idx)

    #   Make the list a numpy array
    number_active = len(active_number_list)
    active_number_list = np.array(active_number_list)

    return number_active, active_number_list


def _iao_complementary_space(iao_ref, s, number_inactive):
    """Determine the complementary space orbitals.

    Args:
        iao_ref (numpy.array): IAO in occupied space.
        s (numpy.array): The overlap matrix.
        number_inactive (int): The number of inactive orbitals.

    Returns:
        numpy.array: The inactive part in IAO (float64).
    """

    #   Construct the "density matrix" for active space
    density_active = np.dot(iao_ref, iao_ref.T)

    #   Get the MO Coefficient from the IAO density matrix
    a_mat = reduce(np.dot, (s, density_active, s))
    eigval, eigvec = scipy.linalg.eigh(a=a_mat, b=s)

    #   Extract inactive part of "MO Coefficient" and return it
    eigen_vectors = eigvec[:, : number_inactive]

    return eigen_vectors


def _iao_atoms(mol, iao1, iao2):
    """Assign IAO to atom centers and rearrange the IAOs.

    Args:
        mol (pyscf.gto.Mole): The molecule to simulate.
        mf (pyscf.scf.RHF): The mean field of the molecule.
        iao1 (numpy.array): IAO for occupied space (float64).
        iao2 (numpy.array): IAO for complementary space (float64).

    Returns:
        numpy.array: The rearranged IAO (float64).
    """

    # Calclate the integrals for assignment
    number_orbitals = mol.nao_nr()
    r_int1e = mol.intor("cint1e_r_sph", 3)
    iao_combine = np.hstack((iao1, iao2))

    # Calculate atom center for each orbital
    x = np.diag(reduce(np.dot, (iao_combine.T, r_int1e[0], iao_combine)))
    y = np.diag(reduce(np.dot, (iao_combine.T, r_int1e[1], iao_combine)))
    z = np.diag(reduce(np.dot, (iao_combine.T, r_int1e[2], iao_combine)))

    # Align the coordinates
    orbitals_temp = np.vstack((x, y, z))
    orbitals = orbitals_temp.T

    # Assign each orbital to atom center
    atom_list = _dmet_atom_list(mol, orbitals)

    # Prepare the orbital labels
    orb_list = _dmet_orb_list(mol, atom_list)

    # Rearrange the orbitals
    iao_combine = iao_combine[:, orb_list]

    # Orthogonalize the orbitals
    s1 = mol.intor_symmetric("int1e_ovlp")
    iao_combine = np.dot(iao_combine, orth.lowdin(reduce(np.dot, (iao_combine.T, s1, iao_combine))))

    return iao_combine


def _dmet_atom_list(mol, orbitals):
    """Assign IAO to atom centers and rearrange the IAOs.

    Args:
        mol (pyscf.gto.Mole): The molecule to simulate.
        orbitals (numpy.array): Coordinates for the orbital centers (float64).

    Returns:
        list: The list for atom assignment for IAO (int).
    """

    # Initialize the list
    number_orbitals = mol.nao_nr()
    newlist = []

    # Calculate the distance from atom centers and determine the nearest
    for i in range(number_orbitals):
        i_temp = 0
        distance_temp = scipy.linalg.norm(orbitals[i, :] - mol.atom_coord(0))
        for j in range(1, mol.natm):
            distance = scipy.linalg.norm(orbitals[i, :] - mol.atom_coord(j))
            if (distance < distance_temp):
                distance_temp = distance
                i_temp = j
            else:
                pass
        newlist.append(i_temp)

    return newlist


def _dmet_orb_list(mol, atom_list):
    """Rearrange the orbital labels.

    Args:
        mol (pyscf.gto.Mole): The molecule to simulate.
        atom_list (list): Atom list for IAO assignment (int).

    Returns:
        list: The orbital list in new order (int).
    """
    newlist = []
    for i in range(mol.natm):
        for j in range(mol.nao_nr()):
            if (atom_list[j] == i):
                newlist.append(j)

    return newlist



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import numpy as np

from tangelo.problem_decomposition.oniom._helpers.helper_classes import Link
from tangelo.molecule_library import xyz_ethane
from tangelo.toolboxes.molecular_computation.molecule import atom_string_to_list


class ONIOMCappingTest(unittest.TestCase):

    def test_unsupported_string_species(self):
        """Test unsupported built-in chemical group (raise ValueError)."""

        with self.assertRaises(ValueError):
            Link(0, 4, factor=1., species="UNSUPPORTED")

    def test_unsupported_custom_species(self):
        """Test unsupported custom chemical group (raise ValueError)."""

        # No ghost atom (X) as first element.
        not_supported_chem_group = [
            ["C", [ 0.71660,  0.89800,  0.64250]],
            ["I", [ 0.53970,  1.76660, -0.00250]],
            ["Cl", [ 0.48990,  0.00050,  0.05510]],
            ["F", [-0.00780,  0.94520,  1.46400]]
        ]

        with self.assertRaises(ValueError):
            Link(0, 4, factor=1., species=not_supported_chem_group)

    def test_cf3(self):
        """Test capping with CF3 (built-in)."""

        broken_link = Link(0, 4, factor=1., species="CF3")
        new_pos = broken_link.relink(xyz_ethane)

        # First CH3 + capping group.
        xyz = xyz_ethane[:4] + new_pos

        # Those position have been verified with a molecular visualization software.
        ref_capped = atom_string_to_list("""
            C     0.71660    0.89800    0.64250
            H     0.53970    1.76660   -0.00250
            H     0.48990    0.00050    0.05510
            H    -0.00780    0.94520    1.46400
            C     2.15410    0.87450    1.16590
            F     2.20191    1.41954    2.38719
            F     2.94837    1.57115    0.34441
            F     2.59308   -0.38799    1.23257
        """)

        # Every atom must be the same (same order too).
        # Position can be almost equals.
        for i, atom in enumerate(xyz):
            self.assertEqual(atom[0], ref_capped[i][0])

        # Testing only the position of the central atom in the chemical group.
        np.testing.assert_almost_equal(xyz[4][1], ref_capped[4][1], decimal=4)

    def test_custom_species(self):
        """Test capping with a custom chemical group (here -CFICl)."""

        cap = [
            ["X", [ 2.15410,  0.87450,  1.16590]],
            ["C", [ 0.71660,  0.89800,  0.64250]],
            ["I", [ 0.53970,  1.76660, -0.00250]],
            ["Cl", [ 0.48990,  0.00050,  0.05510]],
            ["F", [-0.00780,  0.94520,  1.46400]]
        ]

        broken_link = Link(0, 4, factor=1., species=cap)
        new_pos = broken_link.relink(xyz_ethane)

        # First CH3 + capping group.
        xyz = xyz_ethane[:4] + new_pos

        # Those position have been verified with a molecular visualization software.
        ref_capped = atom_string_to_list("""
            C     0.71660    0.89800    0.64250
            H     0.53970    1.76660   -0.00250
            H     0.48990    0.00050    0.05510
            H    -0.00780    0.94520    1.46400
            C     2.15410    0.87450    1.16590
            I     2.22665    0.32520    2.11184
            Cl    2.83194    0.39157    0.45228
            F     2.53182    1.88810    1.34418
        """)

        # Every atom must be the same (same order too).
        # Position can be almost equals.
        for i, atom in enumerate(xyz):
            self.assertEqual(atom[0], ref_capped[i][0])

        # Testing only the position of the central atom in the chemical group.
        np.testing.assert_almost_equal(xyz[4][1], ref_capped[4][1], decimal=4)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Module to create and manipulate unitary coupled cluster operators for
   open shell systems. This requires the number of alpha and beta electrons
   to be specified.
"""

import itertools

from openfermion.utils import down_index, up_index

from tangelo.toolboxes.operators import FermionOperator


def uccsd_openshell_paramsize(n_spinorbitals, n_alpha_electrons, n_beta_electrons):
    """Determine number of independent amplitudes for open-shell UCCSD
     Args:
        n_spinorbitals(int): Number of spin-orbitals in the system
        n_alpha_electrons(int): Number of alpha electrons in the reference state
        n_beta_electrons(int): Number of beta electrons in the reference state
    Returns:
        The number of unique single amplitudes, double amplitudes
        and the number of single alpha and beta amplitudes, as well as the
        number of double alpha-alpha, beta-beta and alpha-beta amplitudes
    """
    if n_spinorbitals % 2 != 0:
        raise ValueError("The total number of spin-orbitals should be even.")

    # Compute the number of occupied and virtual alpha and beta orbitals
    n_orb_a_b = n_spinorbitals // 2
    n_occ_a = n_alpha_electrons
    n_occ_b = n_beta_electrons
    n_virt_a = n_orb_a_b - n_alpha_electrons
    n_virt_b = n_orb_a_b - n_beta_electrons

    # Calculate the number of alpha single amplitudes
    n_single_a = n_occ_a * n_virt_a

    # Calculate the number of beta single amplitudes
    n_single_b = n_occ_b * n_virt_b

    # Calculate the total number of single amplitudes
    n_single_amplitudes = n_single_a + n_single_b

    # Calculate the number of alpha-alpha double amplitudes
    n_double_aa = n_occ_a * (n_occ_a - 1) * n_virt_a * (n_virt_a - 1) // 4

    # Calculate the number of beta-beta double amplitudes
    n_double_bb = n_occ_b * (n_occ_b - 1) * n_virt_b * (n_virt_b - 1) // 4

    # Calculate the number of alpha-beta double amplitudes
    n_double_ab = n_occ_a * n_occ_b * n_virt_a * n_virt_b

    # Calculate the total number of double amplitudes
    n_double_amplitudes = n_double_aa + n_double_bb + n_double_ab

    return n_single_amplitudes, n_double_amplitudes, n_single_a, n_single_b,\
        n_double_aa, n_double_bb, n_double_ab


def uccsd_openshell_generator(packed_amplitudes, n_spinorbitals, n_alpha_electrons,
                              n_beta_electrons, anti_hermitian=True):
    r"""Create an open-shell UCCSD generator for a system with n_alpha_electrons and
       n_beta_electrons
    This function generates a FermionOperator for a UCCSD generator designed
        to act on a single reference state consisting of n_spinorbitals spin orbitals
        and n_alpha_electrons alpha electrons and n_beta_electrons beta electrons,
        that is a high-spin open-shell operator
    Args:
        packed_amplitudes(list): List storing the unique single
            and double excitation amplitudes for an open-shell UCCSD operator.
            The ordering lists unique single excitations before double
            excitations.
        n_spinorbitals(int): Number of spin-orbitals used to represent the system
        n_alpha_electrons(int): Number of alpha electrons in the physical system.
        n_beta_electrons(int): Number of beta electrons in the physical system.
        anti_hermitian(Bool): Flag to generate only normal CCSD operator
            rather than unitary variant, primarily for testing
    Returns:
        generator(FermionOperator): Generator of the UCCSD operator that
            builds the open-shell UCCSD wavefunction.
    """
    if n_spinorbitals % 2 != 0:
        raise ValueError("The total number of spin-orbitals should be even.")

    # Compute the number of occupied and virtual alpha and beta orbitals
    n_orb_a_b = n_spinorbitals // 2
    n_occ_a = n_alpha_electrons
    n_occ_b = n_beta_electrons
    n_virt_a = n_orb_a_b - n_alpha_electrons
    n_virt_b = n_orb_a_b - n_beta_electrons

    # Unpack the single and double amplitudes
    _, _, n_single_a, n_single_b, \
        n_double_aa, n_double_bb, _ = uccsd_openshell_paramsize(n_spinorbitals, n_alpha_electrons, n_beta_electrons)

    # Define the various increments for the sizes of the orbital spaces
    n_s_1 = n_single_a
    n_s_2 = n_single_a + n_single_b
    n_d_1 = n_s_2 + n_double_aa
    n_d_2 = n_d_1 + n_double_bb

    # Alpha single amplitudes
    t1_a = packed_amplitudes[:n_s_1]
    # Beta single amplitudes
    t1_b = packed_amplitudes[n_s_1:n_s_2]
    # Alpha-Alpha double amplitudes
    t2_aa = packed_amplitudes[n_s_2:n_d_1]
    # Beta-Beta double amplitudes
    t2_bb = packed_amplitudes[n_d_1:n_d_2]
    # Alpha-Beta double amplitudes
    t2_ab = packed_amplitudes[n_d_2:]

    # Initialize operator
    generator = FermionOperator()

    # Generate all spin-conserving single excitations
    # for the alpha spin case
    for i, (p, q) in enumerate(
            itertools.product(range(n_virt_a), range(n_occ_a))):

        # Get indices of spatial orbitals
        virtual_a = n_occ_a + p
        occupied_a = q

        # Map the alpha index to the proper spin orbital index
        virtual_so = up_index(virtual_a)
        occupied_so = up_index(occupied_a)

        # Generate the alpha single excitations
        coeff = t1_a[i]
        generator += FermionOperator((
            (virtual_so, 1),
            (occupied_so, 0)),
            coeff)
        if anti_hermitian:
            generator += FermionOperator((
                (occupied_so, 1),
                (virtual_so, 0)),
                -coeff)

    # Generate all spin-conserving single excitations
    # for the beta spin case
    for i, (p, q) in enumerate(
            itertools.product(range(n_virt_b), range(n_occ_b))):

        # Get indices of spatial orbitals
        virtual_b = n_occ_b + p
        occupied_b = q

        # Map the beta index to the proper spin orbital index
        virtual_so = down_index(virtual_b)
        occupied_so = down_index(occupied_b)

        # Generate the beta single excitations
        coeff = t1_b[i]
        generator += FermionOperator((
            (virtual_so, 1),
            (occupied_so, 0)),
            coeff)
        if anti_hermitian:
            generator += FermionOperator((
                (occupied_so, 1),
                (virtual_so, 0)),
                -coeff)

    # Generate all spin-conserving alpha-alpha double excitations
    for i, ((s, q), (r, p)) in enumerate(
            itertools.product(
                itertools.combinations(range(n_occ_a), 2), itertools.combinations(range(n_virt_a), 2)
                )):

        # Get indices of orbitals
        virtual_1 = n_occ_a + p
        occupied_1 = q
        virtual_2 = n_occ_a + r
        occupied_2 = s

        # Generate double excitations
        coeff = t2_aa[i]

        # Map the occupied and virtual alpha indices to spin-orbital indices
        virtual_1_a = up_index(virtual_1)
        occupied_1_a = up_index(occupied_1)
        virtual_2_a = up_index(virtual_2)
        occupied_2_a = up_index(occupied_2)

        generator += FermionOperator((
            (virtual_1_a, 1),
            (occupied_1_a, 0),
            (virtual_2_a, 1),
            (occupied_2_a, 0)),
            coeff)
        if anti_hermitian:
            generator += FermionOperator((
                (occupied_2_a, 1),
                (virtual_2_a, 0),
                (occupied_1_a, 1),
                (virtual_1_a, 0)),
                -coeff)
    # Generate all spin-conserving beta-beta double excitations
    for i, ((s, q), (r, p)) in enumerate(
            itertools.product(
                itertools.combinations(range(n_occ_b), 2), itertools.combinations(range(n_virt_b), 2)
            )):
        # Get indices of orbitals
        virtual_1 = n_occ_b + p
        occupied_1 = q
        virtual_2 = n_occ_b + r
        occupied_2 = s

        # Generate double excitations
        coeff = t2_bb[i]

        # Map the occupied and virtual alpha indices to spin-orbital indices
        virtual_1_b = down_index(virtual_1)
        occupied_1_b = down_index(occupied_1)
        virtual_2_b = down_index(virtual_2)
        occupied_2_b = down_index(occupied_2)

        generator += FermionOperator((
            (virtual_1_b, 1),
            (occupied_1_b, 0),
            (virtual_2_b, 1),
            (occupied_2_b, 0)),
            coeff)
        if anti_hermitian:
            generator += FermionOperator((
                (occupied_2_b, 1),
                (virtual_2_b, 0),
                (occupied_1_b, 1),
                (virtual_1_b, 0)),
                -coeff)
        # Generate all spin-conserving alpha-beta double excitations
    for i, (p, q, r, s) in enumerate(
            itertools.product(
                range(n_virt_a), range(n_occ_a), range(n_virt_b), range(n_occ_b))):

        # Get indices of orbitals
        virtual_1 = n_occ_a + p
        occupied_1 = q
        virtual_2 = n_occ_b + r
        occupied_2 = s

        # Generate double excitations
        coeff = t2_ab[i]

        # Map the alpha and beta occupied and virtual orbitals to spinorbitals
        virtual_1_a = up_index(virtual_1)
        occupied_1_a = up_index(occupied_1)
        virtual_2_b = down_index(virtual_2)
        occupied_2_b = down_index(occupied_2)

        generator += FermionOperator((
            (virtual_1_a, 1),
            (occupied_1_a, 0),
            (virtual_2_b, 1),
            (occupied_2_b, 0)),
            coeff)
        if anti_hermitian:
            generator += FermionOperator((
                (occupied_2_b, 1),
                (virtual_2_b, 0),
                (occupied_1_a, 1),
                (virtual_1_a, 0)),
                -coeff)
    return generator


def uccsd_openshell_get_packed_amplitudes(alpha_double_amplitudes, beta_double_amplitudes,
                                          alpha_beta_double_amplitudes, n_spinorbitals, n_alpha_electrons,
                                          n_beta_electrons, alpha_single_amplitudes=None,
                                          beta_single_amplitudes=None):
    r"""Convert amplitudes for use with the open-shell UCCSD (e.g. from a UHF MP2 guess)
    The output list contains only the non-redundant amplitudes that are
    relevant to open-shell UCCSD, in an order suitable for use with the function
    `uccsd_openshell_generator`.
    Args:
        alpha_single_amplitudes(ndarray): [N_virtual_alpha x N_occupied_alpha]
            array string the alpha single excitation amplitudes corresponding
            to t[i_alpha,a_alpha] * (a_a_alpha^\dagger a_i_alpha)
        beta_single_amplitudes(ndarray): [N_virtual_beta x N_occupied_beta]
            array string the beta single excitation amplitudes corresponding
            to t[i_beta,a_beta] * (a_a_beta^\dagger a_i_beta)
        alpha_double_amplitudes(ndarray): [N_occupied_alpha x N_occupied_alpha
            x N_virtual_alpha x N_virtual_alpha] array storing the alpha-alpha
            double excitation amplitudes corresponding to
            t[i_alpha,j_alpha,a_alpha,b_alpha] * (a_a_alpha^\dagger a_i_alpha
            a_b_alpha^\dagger a_j_alpha - H.C.)
        beta_double_amplitudes(ndarray): [N_occupied_beta x N_occupied_beta
            x N_virtual_beta x N_virtual_beta] array storing the beta-beta
            double excitation amplitudes corresponding to
            t[i_beta,j_beta,a_beta,b_beta] * (a_a_beta^\dagger a_i_beta
            a_b_beta^\dagger a_j_beta - H.C.)
        alpha_beta_double_amplitudes(ndarray): [N_occupied_alpha x N_occupied_beta
            x N_virtual_alpha x N_virtual_beta] array storing the alpha-beta
            double excitation amplitudes corresponding to
            t[i_alpha,j_beta,a_alpha,b_beta] * (a_a_alpha^\dagger a_i_alpha
            a_b_beta^\dagger a_j_beta - H.C.)
        n_spinorbitals(int): Number of spin-orbitals used to represent the system
        n_alpha_electrons(int): Number of alpha electrons in the physical system.
        n_beta_electrons(int): Number of beta electrons in the physical system
        alpha_single_amplitudes(ndarray optional): optional [N_occupied_alpha
            x N_virtual_alpha] array string the alpha single excitation
            amplitudes corresponding to t[i_alpha,a_alpha]
            * (a_a_alpha^\dagger a_i_alpha)
        beta_single_amplitudes(ndarray optional): optional [N_occupied_beta
            x N_virtual_beta] array string the beta single excitation
            amplitudes corresponding to t[i_beta,a_beta]
            * (a_a_beta^\dagger a_i_beta)
    Returns:
        packed_amplitudes(list): List storing the unique single (zero in UHF-MP2)
            and double excitation amplitudes for an open-shell UCCSD operator.
            The ordering lists unique single excitations before double
            excitations.
    """

    if n_spinorbitals % 2 != 0:
        raise ValueError("The total number of spin-orbitals should be even.")

    # Compute the number of occupied and virtual alpha and beta orbitals
    n_orb_a_b = n_spinorbitals // 2
    n_occ_a = n_alpha_electrons
    n_occ_b = n_beta_electrons
    n_virt_a = n_orb_a_b - n_alpha_electrons
    n_virt_b = n_orb_a_b - n_beta_electrons

    # Calculate the number of non-redundant single and double amplitudes
    _, _, n_single_a, n_single_b, _, _, _ = uccsd_openshell_paramsize(n_spinorbitals, n_alpha_electrons, n_beta_electrons)

    # packed amplitudes list
    packed_amplitudes = []

    #  Extract all the single excitation amplitudes
    #  for the alpha spin case (they are zero for a UHF
    #  reference due to the Brillouin condition)
    if alpha_single_amplitudes is None:
        packed_amplitudes = [0.]*n_single_a
    else:
        for (p, q) in itertools.product(range(n_virt_a), range(n_occ_a)):

            # Get the amplitude
            packed_amplitudes.append(alpha_single_amplitudes[q, p])

    #  Extract all the single excitation amplitudes
    #  for the beta spin case (they are zero for a UHF
    #  reference due to the Brillouin condition)
    if beta_single_amplitudes is None:
        packed_amplitudes += [0.]*n_single_b
    else:
        for (p, q) in itertools.product(range(n_virt_b), range(n_occ_b)):

            # Get the amplitude
            packed_amplitudes.append(beta_single_amplitudes[q, p])

    # Extract all of the non-redundant alpha-alpha double
    # excitation amplitudes
    for ((s, q), (r, p)) in itertools.product(
                itertools.combinations(range(n_occ_a), 2), itertools.combinations(range(n_virt_a), 2)
                ):

        # Get the amplitude
        packed_amplitudes.append(alpha_double_amplitudes[q, s, p, r])

    # Extract all of the non-redundant beta-beta double
    # excitation amplitudes
    for ((s, q), (r, p)) in itertools.product(
                itertools.combinations(range(n_occ_b), 2), itertools.combinations(range(n_virt_b), 2)
                ):

        # Get the amplitude
        packed_amplitudes.append(beta_double_amplitudes[q, s, p, r])

    # Extract all of the non-redundant alpha-beta double
    # excitation amplitudes
    for (p, q, r, s) in itertools.product(
                range(n_virt_a), range(n_occ_a), range(n_virt_b), range(n_occ_b)):

        # Get the amplitude
        packed_amplitudes.append(alpha_beta_double_amplitudes[q, s, p, r])

    return packed_amplitudes



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from tangelo.linq import Simulator
from tangelo.molecule_library import mol_H2_sto3g, mol_H4_sto3g
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping
from tangelo.toolboxes.qubit_mappings.statevector_mapping import get_reference_circuit, get_vector, vector_to_circuit
from tangelo.toolboxes.ansatz_generator.fermionic_operators import number_operator, spinz_operator, spin2_operator

# Initiate simulator
sim = Simulator()


class fermionic_operators_Test(unittest.TestCase):

    def test_number_operator_hf_ref(self):
        """Verify that the number penalty terms return zero for all mappings
        given the correct number of electrons and mapping.
        """

        for mol in [mol_H2_sto3g, mol_H4_sto3g]:
            for mapping in ["jw", "bk", "scbk"]:
                hf_state_circuit = get_reference_circuit(n_spinorbitals=mol.n_active_sos,
                                                         n_electrons=mol.n_active_electrons,
                                                         mapping=mapping,
                                                         up_then_down=True)
                num_op = number_operator(mol.n_active_mos, up_then_down=False)
                qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=num_op,
                                                             mapping=mapping,
                                                             n_spinorbitals=mol.n_active_sos,
                                                             n_electrons=mol.n_active_electrons,
                                                             up_then_down=True)

                # Assert energy returned is as expected for given parameters
                numval = sim.get_expectation_value(qubit_hamiltonian, hf_state_circuit)
                self.assertAlmostEqual(numval, mol.n_active_electrons, delta=1e-6)

    def test_spinz_operator_hf_ref(self):
        """Verify that the number penalty terms return zero for all mappings
        given the correct number of electrons and mapping.
        """

        for mol in [mol_H2_sto3g, mol_H4_sto3g]:
            for mapping in ["jw", "bk", "scbk"]:
                hf_state_circuit = get_reference_circuit(n_spinorbitals=mol.n_active_sos,
                                                         n_electrons=mol.n_active_electrons,
                                                         mapping=mapping,
                                                         up_then_down=True)
                spinz_op = spinz_operator(mol.n_active_mos, up_then_down=False)
                qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=spinz_op,
                                                             mapping=mapping,
                                                             n_spinorbitals=mol.n_active_sos,
                                                             n_electrons=mol.n_active_electrons,
                                                             up_then_down=True)

                # Assert energy returned is as expected for given parameters
                spinzval = sim.get_expectation_value(qubit_hamiltonian, hf_state_circuit)
                self.assertAlmostEqual(spinzval, 0.0, delta=1e-6)

    def test_spin2_penalty_terms_hf_ref(self):
        """Verify that the number penalty terms return zero for all mappings
        given the correct number of S^2 and mapping.
        """

        for mol in [mol_H2_sto3g, mol_H4_sto3g]:
            for mapping in ["jw", "bk", "scbk"]:
                hf_state_circuit = get_reference_circuit(n_spinorbitals=mol.n_active_sos,
                                                         n_electrons=mol.n_active_electrons,
                                                         mapping=mapping,
                                                         up_then_down=True)
                pen_op = spin2_operator(mol.n_active_mos, up_then_down=False)
                qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=pen_op,
                                                             mapping=mapping,
                                                             n_spinorbitals=mol.n_active_sos,
                                                             n_electrons=mol.n_active_electrons,
                                                             up_then_down=True)

                # Assert energy returned is as expected for given parameters
                spin2val = sim.get_expectation_value(qubit_hamiltonian, hf_state_circuit)
                self.assertAlmostEqual(spin2val, 0.0, delta=1e-6)

    def test_doublet(self):
        """Verify that the number penalty terms return zero for all mappings
        given the correct number of S^2 and mapping.
        """

        for mol in [mol_H2_sto3g, mol_H4_sto3g]:
            for mapping in ["jw"]:  # may add tests for other mappings later
                vector = get_vector(n_spinorbitals=mol.n_active_sos,
                                    n_electrons=mol.n_active_electrons,
                                    mapping=mapping,
                                    up_then_down=True)

                # Add electron with spin up
                vector[mol.n_active_electrons//2] = 1

                # Create doublet state prep circuit
                doublet_state_circuit = vector_to_circuit(vector)

                # Test number of electrons is mol.n_active_electrons+1
                num_op = number_operator(mol.n_active_mos, up_then_down=False)
                qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=num_op,
                                                             mapping=mapping,
                                                             n_spinorbitals=mol.n_active_sos,
                                                             n_electrons=mol.n_active_electrons,
                                                             up_then_down=True)

                # Assert penalty is zero
                numval = sim.get_expectation_value(qubit_hamiltonian, doublet_state_circuit)
                self.assertAlmostEqual(numval, mol.n_active_electrons + 1, delta=1e-6)

                # Test that spin of 0.5 returns zero
                spinz_op = spinz_operator(mol.n_active_mos, up_then_down=False)
                qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=spinz_op,
                                                             mapping=mapping,
                                                             n_spinorbitals=mol.n_active_sos,
                                                             n_electrons=mol.n_active_electrons,
                                                             up_then_down=True)

                # Assert penalty value returned is zero
                spinzval = sim.get_expectation_value(qubit_hamiltonian, doublet_state_circuit)
                self.assertAlmostEqual(spinzval, 0.5, delta=1e-6)

                # Test s2 penalty for s2=(0.5)*(0.5+1)=0.75 is zero
                spin2_op = spin2_operator(mol.n_active_mos, up_then_down=False)
                qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=spin2_op,
                                                             mapping=mapping,
                                                             n_spinorbitals=mol.n_active_sos,
                                                             n_electrons=mol.n_active_electrons,
                                                             up_then_down=True)

                # Assert penalty returned is zero
                spin2val = sim.get_expectation_value(qubit_hamiltonian, doublet_state_circuit)
                self.assertAlmostEqual(spin2val, 0.5 * (0.5 + 1), delta=1e-6)

    def test_triplet(self):
        """Verify that the number penalty terms return zero for all mappings
        given the correct number of S^2 and mapping.
        """

        for mol in [mol_H2_sto3g, mol_H4_sto3g]:
            for mapping in ["jw"]:
                vector = get_vector(n_spinorbitals=mol.n_active_sos,
                                    n_electrons=mol.n_active_electrons,
                                    mapping=mapping,
                                    up_then_down=True)

                # Add electron with spin up and remove electron with spin down
                vector[mol.n_active_electrons//2] = 1
                vector[mol.n_active_sos//2 + mol.n_active_electrons//2-1] = 0

                # Create triplet state circuit
                triplet_state_circuit = vector_to_circuit(vector)

                # Test number operator
                num_op = number_operator(mol.n_active_mos, up_then_down=False)
                qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=num_op,
                                                             mapping=mapping,
                                                             n_spinorbitals=mol.n_active_sos,
                                                             n_electrons=mol.n_active_electrons,
                                                             up_then_down=True)

                # Assert penalty is zero
                numval = sim.get_expectation_value(qubit_hamiltonian, triplet_state_circuit)
                self.assertAlmostEqual(numval, mol.n_active_electrons, delta=1e-6)

                # Test sz value, should be 1 for triplet state
                spinz_op = spinz_operator(mol.n_active_mos, up_then_down=False)
                qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=spinz_op,
                                                             mapping=mapping,
                                                             n_spinorbitals=mol.n_active_sos,
                                                             n_electrons=mol.n_active_electrons,
                                                             up_then_down=True)

                # Assert penalty term is zero
                spinzval = sim.get_expectation_value(qubit_hamiltonian, triplet_state_circuit)
                self.assertAlmostEqual(spinzval, 1, delta=1e-6)

                # Test S^2 penalty, S2 value shoule be (1)*(1+1)=2 for a triplet state
                spin2_op = spin2_operator(mol.n_active_mos, up_then_down=False)
                qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=spin2_op,
                                                             mapping=mapping,
                                                             n_spinorbitals=mol.n_active_sos,
                                                             n_electrons=mol.n_active_electrons,
                                                             up_then_down=True)

                # Assert penalty returns zero
                spin2val = sim.get_expectation_value(qubit_hamiltonian, triplet_state_circuit)
                self.assertAlmostEqual(spin2val, 1*(1 + 1), delta=1e-6)


if __name__ == "__main__":
    unittest.main()
